import os
import re
import weakref

import numpy as np

from yt.data_objects.index_subobjects.grid_patch import AMRGridPatch
from yt.data_objects.static_output import Dataset
from yt.fields.field_info_container import FieldInfoContainer
from yt.funcs import mylog, setdefaultattr
from yt.geometry.api import Geometry
from yt.geometry.grid_geometry_handler import GridIndex
from yt.utilities.file_handler import HDF5FileHandler, valid_hdf5_signature
from yt.utilities.lib.misc_utilities import get_box_grids_level
from yt.utilities.on_demand_imports import _h5py as h5py
from yt.utilities.parallel_tools.parallel_analysis_interface import parallel_root_only

from .fields import (
    ChomboFieldInfo,
    ChomboPICFieldInfo1D,
    ChomboPICFieldInfo2D,
    ChomboPICFieldInfo3D,
    Orion2FieldInfo,
    PlutoFieldInfo,
)


def is_chombo_hdf5(fn):
    try:
        with h5py.File(fn, mode="r") as fileh:
            valid = "Chombo_global" in fileh["/"]
    except (KeyError, OSError, ImportError):
        return False
    return valid


class ChomboGrid(AMRGridPatch):
    _id_offset = 0
    __slots__ = ["_level_id", "stop_index"]

    def __init__(self, id, index, level, start, stop):
        AMRGridPatch.__init__(self, id, filename=index.index_filename, index=index)
        self._parent_id = []
        self._children_ids = []
        self.Level = level
        self.ActiveDimensions = stop - start + 1

    def get_global_startindex(self):
        """
        Return the integer starting index for each dimension at the current
        level.

        """
        if self.start_index is not None:
            return self.start_index
        if self.Parent is None:
            iLE = self.LeftEdge - self.ds.domain_left_edge
            start_index = iLE / self.dds
            return np.rint(start_index).astype("int64").ravel()
        pdx = self.Parent[0].dds
        start_index = (self.Parent[0].get_global_startindex()) + np.rint(
            (self.LeftEdge - self.Parent[0].LeftEdge) / pdx
        )
        self.start_index = (start_index * self.ds.refine_by).astype("int64").ravel()
        return self.start_index

    def _setup_dx(self):
        # has already been read in and stored in index
        self.dds = self.ds.arr(self.index.dds_list[self.Level], "code_length")

    @property
    def Parent(self):
        if len(self._parent_id) == 0:
            return None
        return [self.index.grids[pid - self._id_offset] for pid in self._parent_id]

    @property
    def Children(self):
        return [self.index.grids[cid - self._id_offset] for cid in self._children_ids]


class ChomboHierarchy(GridIndex):
    grid = ChomboGrid
    _data_file = None

    def __init__(self, ds, dataset_type="chombo_hdf5"):
        self.domain_left_edge = ds.domain_left_edge
        self.domain_right_edge = ds.domain_right_edge
        self.dataset_type = dataset_type
        self.field_indexes = {}
        self.dataset = weakref.proxy(ds)
        # for now, the index file is the dataset!
        self.index_filename = os.path.abspath(self.dataset.parameter_filename)
        self.directory = ds.directory
        self._handle = ds._handle

        self._levels = [key for key in self._handle.keys() if key.startswith("level")]

        GridIndex.__init__(self, ds, dataset_type)

        self._read_particles()

    def _read_particles(self):
        # only do anything if the dataset contains particles
        if not any(f[1].startswith("particle_") for f in self.field_list):
            return

        self.num_particles = 0
        particles_per_grid = []
        for key, val in self._handle.items():
            if key.startswith("level"):
                level_particles = val["particles:offsets"][:]
                self.num_particles += level_particles.sum()
                particles_per_grid = np.concatenate(
                    (particles_per_grid, level_particles)
                )

        for i, _grid in enumerate(self.grids):
            self.grids[i].NumberOfParticles = particles_per_grid[i]
            self.grid_particle_count[i] = particles_per_grid[i]

        assert self.num_particles == self.grid_particle_count.sum()

    # Chombo datasets, by themselves, have no "known" fields. However,
    # we will look for "fluid" fields by finding the string "component" in
    # the output file, and "particle" fields by finding the string "particle".
    def _detect_output_fields(self):
        # look for fluid fields
        output_fields = []
        for key, val in self._handle.attrs.items():
            if key.startswith("component"):
                output_fields.append(val.decode("ascii"))
        self.field_list = [("chombo", c) for c in output_fields]

        # look for particle fields
        particle_fields = []
        for key, val in self._handle.attrs.items():
            if key.startswith("particle"):
                particle_fields.append(val.decode("ascii"))
        self.field_list.extend([("io", c) for c in particle_fields])

    def _count_grids(self):
        self.num_grids = 0
        for lev in self._levels:
            d = self._handle[lev]
            if "Processors" in d:
                self.num_grids += d["Processors"].len()
            elif "boxes" in d:
                self.num_grids += d["boxes"].len()
            else:
                raise RuntimeError("Unknown file specification")

    def _parse_index(self):
        f = self._handle  # shortcut
        self.max_level = f.attrs["num_levels"] - 1

        grids = []
        self.dds_list = []
        i = 0
        D = self.dataset.dimensionality
        for lev_index, lev in enumerate(self._levels):
            level_number = int(re.match(r"level_(\d+)", lev).groups()[0])
            try:
                boxes = f[lev]["boxes"][()]
            except KeyError:
                boxes = f[lev]["particles:boxes"][()]
            dx = f[lev].attrs["dx"]
            self.dds_list.append(dx * np.ones(3))

            if D == 1:
                self.dds_list[lev_index][1] = 1.0
                self.dds_list[lev_index][2] = 1.0

            if D == 2:
                self.dds_list[lev_index][2] = 1.0

            for level_id, box in enumerate(boxes):
                si = np.array([box[f"lo_{ax}"] for ax in "ijk"[:D]])
                ei = np.array([box[f"hi_{ax}"] for ax in "ijk"[:D]])

                if D == 1:
                    si = np.concatenate((si, [0.0, 0.0]))
                    ei = np.concatenate((ei, [0.0, 0.0]))

                if D == 2:
                    si = np.concatenate((si, [0.0]))
                    ei = np.concatenate((ei, [0.0]))

                pg = self.grid(len(grids), self, level=level_number, start=si, stop=ei)
                grids.append(pg)
                grids[-1]._level_id = level_id
                self.grid_levels[i] = level_number
                self.grid_left_edge[i] = self.dds_list[lev_index] * si.astype(
                    self.float_type
                )
                self.grid_right_edge[i] = self.dds_list[lev_index] * (
                    ei.astype(self.float_type) + 1
                )
                self.grid_particle_count[i] = 0
                self.grid_dimensions[i] = ei - si + 1
                i += 1
        self.grids = np.empty(len(grids), dtype="object")
        for gi, g in enumerate(grids):
            self.grids[gi] = g

    def _populate_grid_objects(self):
        self._reconstruct_parent_child()
        for g in self.grids:
            g._prepare_grid()
            g._setup_dx()

    def _setup_derived_fields(self):
        self.derived_field_list = []

    def _reconstruct_parent_child(self):
        mask = np.empty(len(self.grids), dtype="int32")
        mylog.debug("First pass; identifying child grids")
        for i, grid in enumerate(self.grids):
            get_box_grids_level(
                self.grid_left_edge[i, :],
                self.grid_right_edge[i, :],
                self.grid_levels[i].item() + 1,
                self.grid_left_edge,
                self.grid_right_edge,
                self.grid_levels,
                mask,
            )
            ids = np.where(mask.astype("bool"))  # where is a tuple
            grid._children_ids = ids[0] + grid._id_offset
        mylog.debug("Second pass; identifying parents")
        for i, grid in enumerate(self.grids):  # Second pass
            for child in grid.Children:
                child._parent_id.append(i + grid._id_offset)


class ChomboDataset(Dataset):
    _load_requirements = ["h5py"]
    _index_class = ChomboHierarchy
    _field_info_class: type[FieldInfoContainer] = ChomboFieldInfo

    def __init__(
        self,
        filename,
        dataset_type="chombo_hdf5",
        storage_filename=None,
        ini_filename=None,
        units_override=None,
        unit_system="cgs",
        default_species_fields=None,
    ):
        self.fluid_types += ("chombo",)
        self._handle = HDF5FileHandler(filename)
        self.dataset_type = dataset_type

        self.geometry = Geometry.CARTESIAN
        self.ini_filename = ini_filename
        self.fullplotdir = os.path.abspath(filename)
        Dataset.__init__(
            self,
            filename,
            self.dataset_type,
            units_override=units_override,
            unit_system=unit_system,
            default_species_fields=default_species_fields,
        )
        self.storage_filename = storage_filename
        self.cosmological_simulation = False

        # These are parameters that I very much wish to get rid of.
        self.parameters["HydroMethod"] = "chombo"
        self.parameters["DualEnergyFormalism"] = 0
        self.parameters["EOSType"] = -1  # default

    def _set_code_unit_attributes(self):
        if not hasattr(self, "length_unit"):
            mylog.warning("Setting code length unit to be 1.0 cm")
        if not hasattr(self, "mass_unit"):
            mylog.warning("Setting code mass unit to be 1.0 g")
        if not hasattr(self, "time_unit"):
            mylog.warning("Setting code time unit to be 1.0 s")
        setdefaultattr(self, "length_unit", self.quan(1.0, "cm"))
        setdefaultattr(self, "mass_unit", self.quan(1.0, "g"))
        setdefaultattr(self, "time_unit", self.quan(1.0, "s"))
        setdefaultattr(self, "magnetic_unit", self.quan(np.sqrt(4.0 * np.pi), "gauss"))
        setdefaultattr(self, "velocity_unit", self.length_unit / self.time_unit)

    def _localize(self, f, default):
        if f is None:
            return os.path.join(self.directory, default)
        return f

    def _parse_parameter_file(self):
        self.dimensionality = self._handle["Chombo_global/"].attrs["SpaceDim"]
        self.domain_left_edge = self._calc_left_edge()
        self.domain_right_edge = self._calc_right_edge()
        self.domain_dimensions = self._calc_domain_dimensions()

        # if a lower-dimensional dataset, set up pseudo-3D stuff here.
        if self.dimensionality == 1:
            self.domain_left_edge = np.concatenate((self.domain_left_edge, [0.0, 0.0]))
            self.domain_right_edge = np.concatenate(
                (self.domain_right_edge, [1.0, 1.0])
            )
            self.domain_dimensions = np.concatenate((self.domain_dimensions, [1, 1]))

        if self.dimensionality == 2:
            self.domain_left_edge = np.concatenate((self.domain_left_edge, [0.0]))
            self.domain_right_edge = np.concatenate((self.domain_right_edge, [1.0]))
            self.domain_dimensions = np.concatenate((self.domain_dimensions, [1]))

        self.refine_by = self._handle["/level_0"].attrs["ref_ratio"]
        self._determine_periodic()
        self._determine_current_time()

    def _determine_current_time(self):
        # some datasets will not be time-dependent, and to make
        # matters worse, the simulation time is not always
        # stored in the same place in the hdf file! Make
        # sure we handle that here.
        try:
            self.current_time = self._handle.attrs["time"]
        except KeyError:
            try:
                self.current_time = self._handle["/level_0"].attrs["time"]
            except KeyError:
                self.current_time = 0.0

    def _determine_periodic(self):
        # we default to true unless the HDF5 file says otherwise
        is_periodic = np.array([True, True, True])
        for dir in range(self.dimensionality):
            try:
                is_periodic[dir] = self._handle["/level_0"].attrs[
                    "is_periodic_%d" % dir
                ]
            except KeyError:
                is_periodic[dir] = True
        self._periodicity = tuple(is_periodic)

    def _calc_left_edge(self):
        fileh = self._handle
        dx0 = fileh["/level_0"].attrs["dx"]
        D = self.dimensionality
        LE = dx0 * ((np.array(list(fileh["/level_0"].attrs["prob_domain"])))[0:D])
        return LE

    def _calc_right_edge(self):
        fileh = self._handle
        dx0 = fileh["/level_0"].attrs["dx"]
        D = self.dimensionality
        RE = dx0 * ((np.array(list(fileh["/level_0"].attrs["prob_domain"])))[D:] + 1)
        return RE

    def _calc_domain_dimensions(self):
        fileh = self._handle
        D = self.dimensionality
        L_index = (np.array(list(fileh["/level_0"].attrs["prob_domain"])))[0:D]
        R_index = (np.array(list(fileh["/level_0"].attrs["prob_domain"])))[D:] + 1
        return R_index - L_index

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        if cls._missing_load_requirements():
            return False

        if not is_chombo_hdf5(filename):
            return False

        pluto_ini_file_exists = False
        orion2_ini_file_exists = False

        if isinstance(filename, str):
            dir_name = os.path.dirname(os.path.abspath(filename))
            pluto_ini_filename = os.path.join(dir_name, "pluto.ini")
            orion2_ini_filename = os.path.join(dir_name, "orion2.ini")
            pluto_ini_file_exists = os.path.isfile(pluto_ini_filename)
            orion2_ini_file_exists = os.path.isfile(orion2_ini_filename)

        if not (pluto_ini_file_exists or orion2_ini_file_exists):
            try:
                fileh = h5py.File(filename, mode="r")
                valid = "Chombo_global" in fileh["/"]
                # ORION2 simulations should always have this:
                valid = valid and "CeilVA_mass" not in fileh.attrs.keys()
                valid = valid and "Charm_global" not in fileh.keys()
                fileh.close()
                return valid
            except Exception:
                pass
        return False

    @parallel_root_only
    def print_key_parameters(self):
        for a in [
            "current_time",
            "domain_dimensions",
            "domain_left_edge",
            "domain_right_edge",
        ]:
            if not hasattr(self, a):
                mylog.error("Missing %s in parameter file definition!", a)
                continue
            v = getattr(self, a)
            mylog.info("Parameters: %-25s = %s", a, v)


class PlutoHierarchy(ChomboHierarchy):
    def __init__(self, ds, dataset_type="chombo_hdf5"):
        ChomboHierarchy.__init__(self, ds, dataset_type)

    def _parse_index(self):
        f = self._handle  # shortcut
        self.max_level = f.attrs["num_levels"] - 1

        grids = []
        self.dds_list = []
        i = 0
        D = self.dataset.dimensionality
        for lev_index, lev in enumerate(self._levels):
            level_number = int(re.match(r"level_(\d+)", lev).groups()[0])
            try:
                boxes = f[lev]["boxes"][()]
            except KeyError:
                boxes = f[lev]["particles:boxes"][()]
            dx = f[lev].attrs["dx"]
            self.dds_list.append(dx * np.ones(3))

            if D == 1:
                self.dds_list[lev_index][1] = 1.0
                self.dds_list[lev_index][2] = 1.0

            if D == 2:
                self.dds_list[lev_index][2] = 1.0

            for level_id, box in enumerate(boxes):
                si = np.array([box[f"lo_{ax}"] for ax in "ijk"[:D]])
                ei = np.array([box[f"hi_{ax}"] for ax in "ijk"[:D]])

                if D == 1:
                    si = np.concatenate((si, [0.0, 0.0]))
                    ei = np.concatenate((ei, [0.0, 0.0]))

                if D == 2:
                    si = np.concatenate((si, [0.0]))
                    ei = np.concatenate((ei, [0.0]))

                pg = self.grid(len(grids), self, level=level_number, start=si, stop=ei)
                grids.append(pg)
                grids[-1]._level_id = level_id
                self.grid_levels[i] = level_number
                self.grid_left_edge[i] = (
                    self.dds_list[lev_index] * si.astype(self.float_type)
                    + self.domain_left_edge.value
                )
                self.grid_right_edge[i] = (
                    self.dds_list[lev_index] * (ei.astype(self.float_type) + 1)
                    + self.domain_left_edge.value
                )
                self.grid_particle_count[i] = 0
                self.grid_dimensions[i] = ei - si + 1
                i += 1
        self.grids = np.empty(len(grids), dtype="object")
        for gi, g in enumerate(grids):
            self.grids[gi] = g


class PlutoDataset(ChomboDataset):
    _index_class = PlutoHierarchy
    _field_info_class = PlutoFieldInfo

    def __init__(
        self,
        filename,
        dataset_type="chombo_hdf5",
        storage_filename=None,
        ini_filename=None,
        units_override=None,
        unit_system="cgs",
        default_species_fields=None,
    ):
        ChomboDataset.__init__(
            self,
            filename,
            dataset_type,
            storage_filename,
            ini_filename,
            units_override=units_override,
            unit_system=unit_system,
            default_species_fields=default_species_fields,
        )

    def _parse_parameter_file(self):
        """
        Check to see whether a 'pluto.ini' file
        exists in the plot file directory. If one does, attempt to parse it.
        Otherwise grab the dimensions from the hdf5 file.
        """

        pluto_ini_file_exists = False
        dir_name = os.path.dirname(os.path.abspath(self.fullplotdir))
        pluto_ini_filename = os.path.join(dir_name, "pluto.ini")
        pluto_ini_file_exists = os.path.isfile(pluto_ini_filename)

        self.dimensionality = self._handle["Chombo_global/"].attrs["SpaceDim"]
        self.domain_dimensions = self._calc_domain_dimensions()
        self.refine_by = self._handle["/level_0"].attrs["ref_ratio"]

        if pluto_ini_file_exists:
            lines = [line.strip() for line in open(pluto_ini_filename)]
            domain_left_edge = np.zeros(self.dimensionality)
            domain_right_edge = np.zeros(self.dimensionality)
            for il, ll in enumerate(
                lines[
                    lines.index("[Grid]") + 2 : lines.index("[Grid]")
                    + 2
                    + self.dimensionality
                ]
            ):
                domain_left_edge[il] = float(ll.split()[2])
                domain_right_edge[il] = float(ll.split()[-1])
            self._periodicity = [0] * 3
            for il, ll in enumerate(
                lines[
                    lines.index("[Boundary]") + 2 : lines.index("[Boundary]")
                    + 2
                    + 6 : 2
                ]
            ):
                self.periodicity[il] = ll.split()[1] == "periodic"
            self._periodicity = tuple(self.periodicity)
            for ll in lines[lines.index("[Parameters]") + 2 :]:
                if ll.split()[0] == "GAMMA":
                    self.gamma = float(ll.split()[1])
            self.domain_left_edge = domain_left_edge
            self.domain_right_edge = domain_right_edge
        else:
            self.domain_left_edge = self._calc_left_edge()
            self.domain_right_edge = self._calc_right_edge()
            self._periodicity = (True, True, True)

        # if a lower-dimensional dataset, set up pseudo-3D stuff here.
        if self.dimensionality == 1:
            self.domain_left_edge = np.concatenate((self.domain_left_edge, [0.0, 0.0]))
            self.domain_right_edge = np.concatenate(
                (self.domain_right_edge, [1.0, 1.0])
            )
            self.domain_dimensions = np.concatenate((self.domain_dimensions, [1, 1]))

        if self.dimensionality == 2:
            self.domain_left_edge = np.concatenate((self.domain_left_edge, [0.0]))
            self.domain_right_edge = np.concatenate((self.domain_right_edge, [1.0]))
            self.domain_dimensions = np.concatenate((self.domain_dimensions, [1]))

        self._determine_current_time()

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        if cls._missing_load_requirements():
            return False

        if not is_chombo_hdf5(filename):
            return False

        pluto_ini_file_exists = False

        if isinstance(filename, str):
            dir_name = os.path.dirname(os.path.abspath(filename))
            pluto_ini_filename = os.path.join(dir_name, "pluto.ini")
            pluto_ini_file_exists = os.path.isfile(pluto_ini_filename)

        if pluto_ini_file_exists:
            return True

        return False


class Orion2Hierarchy(ChomboHierarchy):
    def __init__(self, ds, dataset_type="orion_chombo_native"):
        ChomboHierarchy.__init__(self, ds, dataset_type)

    def _detect_output_fields(self):
        # look for fluid fields
        output_fields = []
        for key, val in self._handle.attrs.items():
            if key.startswith("component"):
                output_fields.append(val.decode("ascii"))
        self.field_list = [("chombo", c) for c in output_fields]

        # look for particle fields
        self.particle_filename = self.index_filename[:-4] + "sink"
        if not os.path.exists(self.particle_filename):
            return
        pfield_list = [("io", str(c)) for c in self.io.particle_field_index.keys()]
        self.field_list.extend(pfield_list)

    def _read_particles(self):
        if not os.path.exists(self.particle_filename):
            return
        with open(self.particle_filename) as f:
            lines = f.readlines()
            self.num_stars = int(lines[0].strip().split(" ")[0])
            for num, line in enumerate(lines[1:]):
                particle_position_x = float(line.split(" ")[1])
                particle_position_y = float(line.split(" ")[2])
                particle_position_z = float(line.split(" ")[3])
                coord = [particle_position_x, particle_position_y, particle_position_z]
                # for each particle, determine which grids contain it
                # copied from object_finding_mixin.py
                mask = np.ones(self.num_grids)
                for i in range(len(coord)):
                    np.choose(
                        np.greater(self.grid_left_edge.d[:, i], coord[i]),
                        (mask, 0),
                        mask,
                    )
                    np.choose(
                        np.greater(self.grid_right_edge.d[:, i], coord[i]),
                        (0, mask),
                        mask,
                    )
                ind = np.where(mask == 1)
                selected_grids = self.grids[ind]
                # in orion, particles always live on the finest level.
                # so, we want to assign the particle to the finest of
                # the grids we just found
                if len(selected_grids) != 0:
                    grid = sorted(selected_grids, key=lambda grid: grid.Level)[-1]
                    ind = np.where(self.grids == grid)[0][0]
                    self.grid_particle_count[ind] += 1
                    self.grids[ind].NumberOfParticles += 1

                    # store the position in the *.sink file for fast access.
                    try:
                        self.grids[ind]._particle_line_numbers.append(num + 1)
                    except AttributeError:
                        self.grids[ind]._particle_line_numbers = [num + 1]


class Orion2Dataset(ChomboDataset):
    _load_requirements = ["h5py"]
    _index_class = Orion2Hierarchy
    _field_info_class = Orion2FieldInfo

    def __init__(
        self,
        filename,
        dataset_type="orion_chombo_native",
        storage_filename=None,
        ini_filename=None,
        units_override=None,
        default_species_fields=None,
    ):
        ChomboDataset.__init__(
            self,
            filename,
            dataset_type,
            storage_filename,
            ini_filename,
            units_override=units_override,
            default_species_fields=default_species_fields,
        )

    def _parse_parameter_file(self):
        """
        Check to see whether an 'orion2.ini' file
        exists in the plot file directory. If one does, attempt to parse it.
        Otherwise grab the dimensions from the hdf5 file.
        """

        orion2_ini_file_exists = False
        dir_name = os.path.dirname(os.path.abspath(self.fullplotdir))
        orion2_ini_filename = os.path.join(dir_name, "orion2.ini")
        orion2_ini_file_exists = os.path.isfile(orion2_ini_filename)

        if orion2_ini_file_exists:
            self._parse_inputs_file("orion2.ini")
        self.dimensionality = 3
        self.domain_left_edge = self._calc_left_edge()
        self.domain_right_edge = self._calc_right_edge()
        self.domain_dimensions = self._calc_domain_dimensions()
        self.refine_by = self._handle["/level_0"].attrs["ref_ratio"]
        self._determine_periodic()
        self._determine_current_time()

    def _parse_inputs_file(self, ini_filename):
        self.fullplotdir = os.path.abspath(self.parameter_filename)
        self.ini_filename = self._localize(self.ini_filename, ini_filename)
        lines = open(self.ini_filename).readlines()
        # read the file line by line, storing important parameters
        for line in lines:
            try:
                param, sep, vals = line.partition("=")
                if not sep:
                    # No = sign present, so split by space instead
                    param, sep, vals = line.partition(" ")
                param = param.strip()
                vals = vals.strip()
                if not param:  # skip blank lines
                    continue
                if param[0] == "#":  # skip comment lines
                    continue
                if param[0] == "[":  # skip stanza headers
                    continue
                vals = vals.partition("#")[0]  # strip trailing comments
                try:
                    self.parameters[param] = np.int64(vals)
                except ValueError:
                    try:
                        self.parameters[param] = np.float64(vals)
                    except ValueError:
                        self.parameters[param] = vals
            except ValueError:
                mylog.error("ValueError: '%s'", line)
            if param == "GAMMA":
                self.gamma = np.float64(vals)

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        if cls._missing_load_requirements():
            return False

        if not is_chombo_hdf5(filename):
            return False

        pluto_ini_file_exists = False
        orion2_ini_file_exists = False

        if isinstance(filename, str):
            dir_name = os.path.dirname(os.path.abspath(filename))
            pluto_ini_filename = os.path.join(dir_name, "pluto.ini")
            orion2_ini_filename = os.path.join(dir_name, "orion2.ini")
            pluto_ini_file_exists = os.path.isfile(pluto_ini_filename)
            orion2_ini_file_exists = os.path.isfile(orion2_ini_filename)

        if orion2_ini_file_exists:
            return True

        if not pluto_ini_file_exists:
            try:
                fileh = h5py.File(filename, mode="r")
                valid = "CeilVA_mass" in fileh.attrs.keys()
                valid = (
                    "Chombo_global" in fileh["/"] and "Charm_global" not in fileh["/"]
                )
                valid = valid and "CeilVA_mass" in fileh.attrs.keys()
                fileh.close()
                return valid
            except Exception:
                pass
        return False


class ChomboPICHierarchy(ChomboHierarchy):
    def __init__(self, ds, dataset_type="chombo_hdf5"):
        ChomboHierarchy.__init__(self, ds, dataset_type)


class ChomboPICDataset(ChomboDataset):
    _load_requirements = ["h5py"]
    _index_class = ChomboPICHierarchy
    _field_info_class = ChomboPICFieldInfo3D

    def __init__(
        self,
        filename,
        dataset_type="chombo_hdf5",
        storage_filename=None,
        ini_filename=None,
        units_override=None,
    ):
        ChomboDataset.__init__(
            self,
            filename,
            dataset_type,
            storage_filename,
            ini_filename,
            units_override=units_override,
        )

        if self.dimensionality == 1:
            self._field_info_class = ChomboPICFieldInfo1D

        if self.dimensionality == 2:
            self._field_info_class = ChomboPICFieldInfo2D

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        if not valid_hdf5_signature(filename):
            return False

        if cls._missing_load_requirements():
            return False

        if not is_chombo_hdf5(filename):
            return False

        pluto_ini_file_exists = False
        orion2_ini_file_exists = False

        if isinstance(filename, str):
            dir_name = os.path.dirname(os.path.abspath(filename))
            pluto_ini_filename = os.path.join(dir_name, "pluto.ini")
            orion2_ini_filename = os.path.join(dir_name, "orion2.ini")
            pluto_ini_file_exists = os.path.isfile(pluto_ini_filename)
            orion2_ini_file_exists = os.path.isfile(orion2_ini_filename)

        if orion2_ini_file_exists:
            return False

        if pluto_ini_file_exists:
            return False

        try:
            fileh = h5py.File(filename, mode="r")
            valid = "Charm_global" in fileh["/"]
            fileh.close()
            return valid
        except Exception:
            pass
        return False



import multiprocessing
import os
import sys

import nose
import numpy
import yaml

from yt.config import ytcfg
from yt.utilities.answer_testing.framework import AnswerTesting

numpy.set_printoptions(threshold=5, edgeitems=1, precision=4)


class NoseWorker(multiprocessing.Process):
    def __init__(self, task_queue, result_queue):
        multiprocessing.Process.__init__(self)
        self.task_queue = task_queue
        self.result_queue = result_queue

    def run(self):
        proc_name = self.name
        while True:
            next_task = self.task_queue.get()
            if next_task is None:
                print(f"{proc_name}: Exiting")
                self.task_queue.task_done()
                break
            print(f"{proc_name}: {next_task}")
            result = next_task()
            self.task_queue.task_done()
            self.result_queue.put(result)
            if next_task.exclusive:
                print(f"{proc_name}: Exiting (exclusive)")
                break
        return


class NoseTask:
    def __init__(self, job):
        argv, exclusive = job
        self.argv = argv
        self.name = argv[0]
        self.exclusive = exclusive

    def __call__(self):
        test_dir = ytcfg.get("yt", "test_data_dir")
        answers_dir = os.path.join(test_dir, "answers")
        if "--with-answer-testing" in self.argv and not os.path.isdir(
            os.path.join(answers_dir, self.name)
        ):
            nose.run(
                argv=self.argv + ["--answer-store"],
                addplugins=[AnswerTesting()],
                exit=False,
            )
        if os.path.isfile(f"{self.name}.xml"):
            os.remove(f"{self.name}.xml")
        nose.run(argv=self.argv, addplugins=[AnswerTesting()], exit=False)
        return ""

    def __str__(self):
        return f"WILL DO self.name = {self.name}"


def generate_tasks_input():
    pyver = f"py{sys.version_info.major}{sys.version_info.minor}"
    test_dir = ytcfg.get("yt", "test_data_dir")
    answers_dir = os.path.join(test_dir, "answers")
    tests = yaml.load(open("tests/tests.yaml"), Loader=yaml.FullLoader)

    base_argv = ["-s", "--nologcapture", "--with-xunit"]

    base_answer_argv = [
        f"--local-dir={answers_dir}",
        "--with-answer-testing",
        "--answer-big-data",
        "--local",
    ]

    args = []

    for test in list(tests["other_tests"].keys()):
        args.append(([test] + base_argv + tests["other_tests"][test], True))
    for answer in list(tests["answer_tests"].keys()):
        if tests["answer_tests"][answer] is None:
            continue
        argv = [f"{pyver}_{answer}"]
        argv += base_argv + base_answer_argv
        argv.append(f"--answer-name={argv[0]}")
        argv += tests["answer_tests"][answer]
        args.append((argv, False))

    exclude_answers = []
    answer_tests = tests["answer_tests"]
    for key in answer_tests:
        for t in answer_tests[key]:
            exclude_answers.append(t.replace(".py:", ".").replace("/", "."))
    exclude_answers = [f"--exclude-test={ex}" for ex in exclude_answers]

    args = [
        (
            (item + [f"--xunit-file={item[0]}.xml"], exclusive)
            if item[0] != "unittests"
            else (item + ["--xunit-file=unittests.xml"] + exclude_answers, exclusive)
        )
        for item, exclusive in args
    ]
    return args


if __name__ == "__main__":
    # multiprocessing.log_to_stderr(logging.DEBUG)
    tasks = multiprocessing.JoinableQueue()
    results = multiprocessing.Queue()

    num_consumers = int(os.environ.get("NUM_WORKERS", 6))
    consumers = [NoseWorker(tasks, results) for i in range(num_consumers)]
    for w in consumers:
        w.start()

    num_jobs = 0
    for job in generate_tasks_input():
        if job[1]:
            num_consumers -= 1  # take into account exclusive jobs
        tasks.put(NoseTask(job))
        num_jobs += 1

    for _i in range(num_consumers):
        tasks.put(None)

    tasks.join()

    while num_jobs:
        result = results.get()
        num_jobs -= 1



import numpy as np
from numpy.testing import assert_array_equal, assert_equal

from yt.frontends.enzo_e.api import EnzoEDataset
from yt.frontends.enzo_e.fields import NODAL_FLAGS
from yt.testing import requires_file, requires_module
from yt.utilities.answer_testing.framework import (
    FieldValuesTest,
    PixelizedProjectionValuesTest,
    create_obj,
    data_dir_load,
    requires_ds,
)
from yt.utilities.on_demand_imports import _h5py as h5py

_fields = (
    ("gas", "density"),
    ("gas", "specific_total_energy"),
    ("gas", "velocity_x"),
    ("gas", "velocity_y"),
)

_pfields = (
    ("all", "particle_position_x"),
    ("all", "particle_position_y"),
    ("all", "particle_position_z"),
    ("all", "particle_velocity_x"),
    ("all", "particle_velocity_y"),
    ("all", "particle_velocity_z"),
)

hello_world = "hello-0210/hello-0210.block_list"
ep_cosmo = "ENZOP_DD0140/ENZOP_DD0140.block_list"
orszag_tang = "ENZOE_orszag-tang_0.5/ENZOE_orszag-tang_0.5.block_list"


@requires_module("h5py")
@requires_file(hello_world)
def test_EnzoEDataset():
    assert isinstance(data_dir_load(hello_world), EnzoEDataset)


@requires_module("h5py")
@requires_ds(hello_world)
def test_hello_world():
    ds = data_dir_load(hello_world)

    dso = [None, ("sphere", ("max", (0.25, "unitary")))]
    for dobj_name in dso:
        for field in _fields:
            for axis in [0, 1, 2]:
                for weight_field in [None, ("gas", "density")]:
                    yield PixelizedProjectionValuesTest(
                        hello_world, axis, field, weight_field, dobj_name
                    )
            yield FieldValuesTest(hello_world, field, dobj_name)
        dobj = create_obj(ds, dobj_name)
        s1 = dobj["index", "ones"].sum()
        s2 = sum(mask.sum() for block, mask in dobj.blocks)
        assert_equal(s1, s2)


@requires_module("h5py")
@requires_ds(ep_cosmo)
def test_particle_fields():
    ds = data_dir_load(ep_cosmo)

    dso = [None, ("sphere", ("max", (0.1, "unitary")))]
    for dobj_name in dso:
        for field in _pfields:
            yield FieldValuesTest(ep_cosmo, field, dobj_name, particle_type=True)
        dobj = create_obj(ds, dobj_name)
        s1 = dobj["index", "ones"].sum()
        s2 = sum(mask.sum() for block, mask in dobj.blocks)
        assert_equal(s1, s2)


@requires_module("h5py")
@requires_file(hello_world)
def test_hierarchy():
    ds = data_dir_load(hello_world)

    fh = h5py.File(ds.index.grids[0].filename, mode="r")
    for grid in ds.index.grids:
        assert_array_equal(
            grid.LeftEdge.d, fh[grid.block_name].attrs["enzo_GridLeftEdge"]
        )
        assert_array_equal(ds.index.grid_left_edge[grid.id], grid.LeftEdge)
        assert_array_equal(ds.index.grid_right_edge[grid.id], grid.RightEdge)
        for child in grid.Children:
            assert (child.LeftEdge >= grid.LeftEdge).all()
            assert (child.RightEdge <= grid.RightEdge).all()
            assert_equal(child.Parent.id, grid.id)
    fh.close()


@requires_module("h5py")
@requires_file(ep_cosmo)
def test_critical_density():
    ds = data_dir_load(ep_cosmo)

    c1 = (
        (ds.r["dark", "particle_mass"].sum() + ds.r["gas", "cell_mass"].sum())
        / ds.domain_width.prod()
        / ds.critical_density
    )
    c2 = (
        ds.omega_matter
        * (1 + ds.current_redshift) ** 3
        / (ds.omega_matter * (1 + ds.current_redshift) ** 3 + ds.omega_lambda)
    )

    assert np.abs(c1 - c2) / max(c1, c2) < 1e-3


@requires_module("h5py")
@requires_file(orszag_tang)
def test_face_centered_bfields():
    # this is based on the enzo frontend test, test_face_centered_mhdct_fields
    ds = data_dir_load(orszag_tang)

    ad = ds.all_data()
    assert len(ds.index.grids) == 4

    for field, flag in NODAL_FLAGS.items():
        dims = ds.domain_dimensions
        assert_equal(ad[field].shape, (dims.prod(), 2 * sum(flag)))

        # the x and y domains are each split across 2 blocks. The z domain
        # is all located on a single block
        block_dims = (dims[0] // 2, dims[1] // 2, dims[2])
        for grid in ds.index.grids:
            assert_equal(grid[field].shape, block_dims + (2 * sum(flag),))

    # Average of face-centered fields should be the same as cell-centered field
    assert_array_equal(
        ad["enzoe", "bfieldi_x"].sum(axis=-1) / 2, ad["enzoe", "bfield_x"]
    )
    assert_array_equal(
        ad["enzoe", "bfieldi_y"].sum(axis=-1) / 2, ad["enzoe", "bfield_y"]
    )
    assert_array_equal(
        ad["enzoe", "bfieldi_z"].sum(axis=-1) / 2, ad["enzoe", "bfield_z"]
    )



from importlib.metadata import version
from importlib.util import find_spec

from packaging.version import Version

__all__ = [
    "IS_IPYTHON",
    "IPYWIDGETS_ENABLED",
]

IS_IPYTHON: bool
HAS_IPYWIDGETS_GE_8: bool
IPYWIDGETS_ENABLED: bool


try:
    # this name is only defined if running within ipython/jupyter
    __IPYTHON__  # type: ignore [name-defined] # noqa: B018
except NameError:
    IS_IPYTHON = False
else:
    IS_IPYTHON = True


HAS_IPYWIDGETS_GE_8 = (
    Version(version("ipywidgets")) >= Version("8.0.0")
    if find_spec("ipywidgets") is not None
    else False
)

IPYWIDGETS_ENABLED = IS_IPYTHON and HAS_IPYWIDGETS_GE_8



import numpy as np
from numpy.testing import assert_almost_equal, assert_equal

from yt.testing import fake_sph_grid_ds

n_ref = 4


def test_building_tree():
    """
    Build an octree and make sure correct number of particles
    """
    ds = fake_sph_grid_ds()
    octree = ds.octree(n_ref=n_ref)
    assert octree["index", "x"].shape[0] == 17


def test_sph_interpolation_scatter():
    """
    Generate an octree, perform some SPH interpolation and check with some
    answer testing
    """

    ds = fake_sph_grid_ds(hsml_factor=26.0)
    ds._sph_ptypes = ("io",)

    ds.use_sph_normalization = False

    octree = ds.octree(n_ref=n_ref)

    density = octree["io", "density"]
    answers = np.array(
        [
            1.00434706,
            1.00434706,
            1.00434706,
            1.00434706,
            1.00434706,
            1.00434706,
            1.00434706,
            0.7762907,
            0.89250848,
            0.89250848,
            0.97039088,
            0.89250848,
            0.97039088,
            0.97039088,
            1.01156175,
        ]
    )

    assert_almost_equal(density.d, answers)


def test_sph_interpolation_gather():
    """
    Generate an octree, perform some SPH interpolation and check with some
    answer testing
    """
    ds = fake_sph_grid_ds(hsml_factor=26.0)
    ds.index
    ds._sph_ptypes = ("io",)

    ds.sph_smoothing_style = "gather"
    ds.num_neighbors = 5
    ds.use_sph_normalization = False

    octree = ds.octree(n_ref=n_ref)

    density = octree["io", "density"]
    answers = np.array(
        [
            0.59240874,
            0.59240874,
            0.59240874,
            0.59240874,
            0.59240874,
            0.59240874,
            0.59240874,
            0.10026846,
            0.77014968,
            0.77014968,
            0.96127825,
            0.77014968,
            0.96127825,
            0.96127825,
            1.21183996,
        ]
    )

    assert_almost_equal(density.d, answers)


def test_octree_properties():
    """
    Generate an octree, and test the refinement, depth and sizes of the cells.
    """
    ds = fake_sph_grid_ds()
    octree = ds.octree(n_ref=n_ref)

    depth = octree["index", "depth"]
    depth_ans = np.array([0] + [1] * 8 + [2] * 8, dtype=np.int64)
    assert_equal(depth, depth_ans)

    size_ans = np.zeros((depth.shape[0], 3), dtype=np.float64)
    for i in range(size_ans.shape[0]):
        size_ans[i, :] = (ds.domain_right_edge - ds.domain_left_edge) / 2.0 ** depth[i]

    dx = octree["index", "dx"].d
    assert_almost_equal(dx, size_ans[:, 0])

    dy = octree["index", "dy"].d
    assert_almost_equal(dy, size_ans[:, 1])

    dz = octree["index", "dz"].d
    assert_almost_equal(dz, size_ans[:, 2])

    refined = octree["index", "refined"]
    refined_ans = np.array([True] + [False] * 7 + [True] + [False] * 8, dtype=np.bool_)
    assert_equal(refined, refined_ans)



import matplotlib.pyplot as plt

import yt

# Load the dataset.
ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

# Create a sphere of radius 1 Mpc centered on the max density location.
sp = ds.sphere("max", (1, "Mpc"))

# Calculate and store the bulk velocity for the sphere.
bulk_velocity = sp.quantities.bulk_velocity()
sp.set_field_parameter("bulk_velocity", bulk_velocity)

# Create a 1D profile object for profiles over radius
# and add a velocity profile.
prof = yt.create_profile(
    sp,
    "radius",
    ("gas", "velocity_magnitude"),
    units={"radius": "kpc"},
    extrema={"radius": ((0.1, "kpc"), (1000.0, "kpc"))},
    weight_field=("gas", "mass"),
)

# Create arrays to plot.
radius = prof.x
mean = prof["gas", "velocity_magnitude"]
std = prof.standard_deviation["gas", "velocity_magnitude"]

# Plot the average velocity magnitude.
plt.loglog(radius, mean, label="Mean")
# Plot the standard deviation of the velocity magnitude.
plt.loglog(radius, std, label="Standard Deviation")
plt.xlabel("r [kpc]")
plt.ylabel("v [cm/s]")
plt.legend()

plt.savefig("velocity_profiles.png")



from yt.fields.field_info_container import FieldInfoContainer
from yt.fields.species_fields import add_species_field_by_fraction, setup_species_fields
from yt.frontends.gadget.api import GadgetFieldInfo
from yt.utilities.chemical_formulas import ChemicalFormula
from yt.utilities.physical_ratios import _primordial_mass_fraction

metal_elements = ["He", "C", "N", "O", "Ne", "Mg", "Si", "Fe"]


class ArepoFieldInfo(GadgetFieldInfo):
    def __init__(self, ds, field_list, slice_info=None):
        if ds.cosmological_simulation:
            GFM_SFT_units = "dimensionless"
        else:
            GFM_SFT_units = "code_length/code_velocity"
        self.known_particle_fields += (
            ("GFM_StellarFormationTime", (GFM_SFT_units, ["stellar_age"], None)),
            ("MagneticField", ("code_magnetic", ["particle_magnetic_field"], None)),
            (
                "MagneticFieldDivergence",
                ("code_magnetic/code_length", ["magnetic_field_divergence"], None),
            ),
            ("GFM_CoolingRate", ("erg*cm**3/s", ["cooling_rate"], None)),
            ("GFM_Metallicity", ("", ["metallicity"], None)),
            ("GFM_Metals_00", ("", ["H_fraction"], None)),
            ("GFM_Metals_01", ("", ["He_fraction"], None)),
            ("GFM_Metals_02", ("", ["C_fraction"], None)),
            ("GFM_Metals_03", ("", ["N_fraction"], None)),
            ("GFM_Metals_04", ("", ["O_fraction"], None)),
            ("GFM_Metals_05", ("", ["Ne_fraction"], None)),
            ("GFM_Metals_06", ("", ["Mg_fraction"], None)),
            ("GFM_Metals_07", ("", ["Si_fraction"], None)),
            ("GFM_Metals_08", ("", ["Fe_fraction"], None)),
            ("GFM_StellarPhotometrics_00", ("", ["U_magnitude"], None)),
            ("GFM_StellarPhotometrics_01", ("", ["B_magnitude"], None)),
            ("GFM_StellarPhotometrics_02", ("", ["V_magnitude"], None)),
            ("GFM_StellarPhotometrics_03", ("", ["K_magnitude"], None)),
            ("GFM_StellarPhotometrics_04", ("", ["g_magnitude"], None)),
            ("GFM_StellarPhotometrics_05", ("", ["r_magnitude"], None)),
            ("GFM_StellarPhotometrics_06", ("", ["i_magnitude"], None)),
            ("GFM_StellarPhotometrics_07", ("", ["z_magnitude"], None)),
            (
                "CosmicRaySpecificEnergy",
                ("code_specific_energy", ["specific_cosmic_ray_energy"], None),
            ),
        )
        super().__init__(ds, field_list, slice_info=slice_info)

    def setup_particle_fields(self, ptype, *args, **kwargs):
        FieldInfoContainer.setup_particle_fields(self, ptype)
        if ptype == "PartType0":
            self.setup_gas_particle_fields(ptype)
            setup_species_fields(self, ptype)

    def setup_gas_particle_fields(self, ptype):
        from yt.fields.magnetic_field import setup_magnetic_field_aliases

        super().setup_gas_particle_fields(ptype)

        # Since the AREPO gas "particles" are Voronoi cells, we can
        # define a volume here
        def _volume(field, data):
            return data["gas", "mass"] / data["gas", "density"]

        self.add_field(
            ("gas", "cell_volume"),
            function=_volume,
            sampling_type="local",
            units=self.ds.unit_system["volume"],
        )

        if (ptype, "InternalEnergy") in self.field_list:

            def _pressure(field, data):
                return (
                    (data.ds.gamma - 1.0)
                    * data[ptype, "density"]
                    * data[ptype, "InternalEnergy"]
                )

            self.add_field(
                ("gas", "pressure"),
                function=_pressure,
                sampling_type="local",
                units=self.ds.unit_system["pressure"],
            )

        if (ptype, "GFM_Metals_00") in self.field_list:
            self.nuclei_names = metal_elements
            self.species_names = ["H"] + metal_elements

        if (ptype, "MagneticField") in self.field_list:
            setup_magnetic_field_aliases(self, ptype, "MagneticField")

        if (ptype, "NeutralHydrogenAbundance") in self.field_list:

            def _h_p0_fraction(field, data):
                return (
                    data[ptype, "GFM_Metals_00"]
                    * data[ptype, "NeutralHydrogenAbundance"]
                )

            self.add_field(
                (ptype, "H_p0_fraction"),
                sampling_type="particle",
                function=_h_p0_fraction,
                units="",
            )

            def _h_p1_fraction(field, data):
                return data[ptype, "GFM_Metals_00"] * (
                    1.0 - data[ptype, "NeutralHydrogenAbundance"]
                )

            self.add_field(
                (ptype, "H_p1_fraction"),
                sampling_type="particle",
                function=_h_p1_fraction,
                units="",
            )

            add_species_field_by_fraction(self, ptype, "H_p0")
            add_species_field_by_fraction(self, ptype, "H_p1")

            for species in ["H", "H_p0", "H_p1"]:
                for suf in ["_density", "_number_density"]:
                    field = f"{species}{suf}"
                    self.alias(("gas", field), (ptype, field))

        if (ptype, "ElectronAbundance") in self.field_list:
            # If we have ElectronAbundance but not NeutralHydrogenAbundance,
            # try first to use the H_fraction, but otherwise we assume the
            # cosmic value for hydrogen to generate the H_number_density
            if (ptype, "NeutralHydrogenAbundance") not in self.field_list:
                m_u = self.ds.quan(1.0, "amu").in_cgs()
                A_H = ChemicalFormula("H").weight
                if (ptype, "GFM_Metals_00") in self.field_list:

                    def _h_number_density(field, data):
                        return (
                            data["gas", "density"]
                            * data["gas", "H_fraction"]
                            / (A_H * m_u)
                        )

                else:
                    X_H = _primordial_mass_fraction["H"]

                    def _h_number_density(field, data):
                        return data["gas", "density"] * X_H / (A_H * m_u)

                self.add_field(
                    (ptype, "H_number_density"),
                    sampling_type="particle",
                    function=_h_number_density,
                    units=self.ds.unit_system["number_density"],
                )
                self.alias(("gas", "H_number_density"), (ptype, "H_number_density"))
                self.alias(("gas", "H_nuclei_density"), ("gas", "H_number_density"))

            def _el_number_density(field, data):
                return (
                    data[ptype, "ElectronAbundance"] * data[ptype, "H_number_density"]
                )

            self.add_field(
                (ptype, "El_number_density"),
                sampling_type="particle",
                function=_el_number_density,
                units=self.ds.unit_system["number_density"],
            )
            self.alias(("gas", "El_number_density"), (ptype, "El_number_density"))

        if (ptype, "GFM_CoolingRate") in self.field_list:
            self.alias(("gas", "cooling_rate"), ("PartType0", "cooling_rate"))

            def _cooling_time(field, data):
                nH = data["gas", "H_nuclei_density"]
                dedt = -data["gas", "cooling_rate"] * nH * nH
                e = 1.5 * data["gas", "pressure"]
                return e / dedt

            self.add_field(
                ("gas", "cooling_time"), _cooling_time, sampling_type="local", units="s"
            )

        if (ptype, "CosmicRaySpecificEnergy") in self.field_list:
            self.alias(
                (ptype, "specific_cosmic_ray_energy"),
                ("gas", "specific_cosmic_ray_energy"),
            )

            def _cr_energy_density(field, data):
                return (
                    data["PartType0", "specific_cosmic_ray_energy"]
                    * data["gas", "density"]
                )

            self.add_field(
                ("gas", "cosmic_ray_energy_density"),
                _cr_energy_density,
                sampling_type="local",
                units=self.ds.unit_system["pressure"],
            )

            def _cr_pressure(field, data):
                return (data.ds.gamma_cr - 1.0) * data[
                    "gas", "cosmic_ray_energy_density"
                ]

            self.add_field(
                ("gas", "cosmic_ray_pressure"),
                _cr_pressure,
                sampling_type="local",
                units=self.ds.unit_system["pressure"],
            )



gadget_header_specs = {
    "default": (
        ("Npart", 6, "i"),
        ("Massarr", 6, "d"),
        ("Time", 1, "d"),
        ("Redshift", 1, "d"),
        ("FlagSfr", 1, "i"),
        ("FlagFeedback", 1, "i"),
        ("Nall", 6, "i"),
        ("FlagCooling", 1, "i"),
        ("NumFiles", 1, "i"),
        ("BoxSize", 1, "d"),
        ("Omega0", 1, "d"),
        ("OmegaLambda", 1, "d"),
        ("HubbleParam", 1, "d"),
        ("FlagAge", 1, "i"),
        ("FlagMetals", 1, "i"),
        ("NallHW", 6, "i"),
        ("unused", 16, "i"),
    ),
    "pad32": (("empty", 32, "c"),),
    "pad64": (("empty", 64, "c"),),
    "pad128": (("empty", 128, "c"),),
    "pad256": (("empty", 256, "c"),),
}

gadget_ptype_specs = {"default": ("Gas", "Halo", "Disk", "Bulge", "Stars", "Bndry")}

gadget_field_specs = {
    "default": (
        "Coordinates",
        "Velocities",
        "ParticleIDs",
        "Mass",
        ("InternalEnergy", "Gas"),
        ("Density", "Gas"),
        ("SmoothingLength", "Gas"),
    ),
    "agora_unlv": (
        "Coordinates",
        "Velocities",
        "ParticleIDs",
        "Mass",
        ("InternalEnergy", "Gas"),
        ("Density", "Gas"),
        ("Electron_Number_Density", "Gas"),
        ("HI_NumberDensity", "Gas"),
        ("SmoothingLength", "Gas"),
    ),
    "group0000": (
        "Coordinates",
        "Velocities",
        "ParticleIDs",
        "Mass",
        "Potential",
        ("Temperature", "Gas"),
        ("Density", "Gas"),
        ("ElectronNumberDensity", "Gas"),
        ("HI_NumberDensity", "Gas"),
        ("SmoothingLength", "Gas"),
        ("StarFormationRate", "Gas"),
        ("DelayTime", "Gas"),
        ("FourMetalFractions", ("Gas", "Stars")),
        ("MaxTemperature", ("Gas", "Stars")),
        ("NStarsSpawned", ("Gas", "Stars")),
        ("StellarAge", "Stars"),
    ),
    "magneticum_box2_hr": (
        "Coordinates",
        "Velocities",
        "ParticleIDs",
        "Mass",
        ("InternalEnergy", "Gas"),
        ("Density", "Gas"),
        ("SmoothingLength", "Gas"),
        ("ColdFraction", "Gas"),
        ("Temperature", "Gas"),
        ("StellarAge", ("Stars", "Bndry")),
        "Potential",
        ("InitialMass", "Stars"),
        ("ElevenMetalMasses", ("Gas", "Stars")),
        ("StarFormationRate", "Gas"),
        ("TrueMass", "Bndry"),
        ("AccretionRate", "Bndry"),
    ),
}

gadget_hdf5_ptypes = (
    "PartType0",
    "PartType1",
    "PartType2",
    "PartType3",
    "PartType4",
    "PartType5",
)

SNAP_FORMAT_2_OFFSET = 16

"""
Here we have a dictionary of possible element species defined in Gadget
datasets, keyed by the number of elements. In some cases, these are mass
fractions, in others, they are metals--the context for the dataset will
determine this. The "Ej" key is for the total mass of all elements that
are not explicitly listed.
"""
elem_names_opts = {
    4: ["C", "O", "Si", "Fe"],
    7: ["C", "N", "O", "Mg", "Si", "Fe", "Ej"],
    8: ["He", "C", "O", "Mg", "S", "Si", "Fe", "Ej"],
    11: ["He", "C", "Ca", "O", "N", "Ne", "Mg", "S", "Si", "Fe", "Ej"],
    15: [
        "He",
        "C",
        "Ca",
        "O",
        "N",
        "Ne",
        "Mg",
        "S",
        "Si",
        "Fe",
        "Na",
        "Al",
        "Ar",
        "Ni",
        "Ej",
    ],
}



from . import tests
from .data_structures import (
    ChomboDataset,
    ChomboGrid,
    ChomboHierarchy,
    ChomboPICDataset,
    ChomboPICHierarchy,
    Orion2Dataset,
    Orion2Hierarchy,
    PlutoDataset,
    PlutoHierarchy,
)
from .fields import (
    ChomboFieldInfo,
    ChomboPICFieldInfo1D,
    ChomboPICFieldInfo2D,
    ChomboPICFieldInfo3D,
    Orion2FieldInfo,
    PlutoFieldInfo,
)
from .io import IOHandlerChomboHDF5



from . import tests
from .data_structures import (
    MoabHex8Dataset,
    MoabHex8Hierarchy,
    MoabHex8Mesh,
    PyneMoabHex8Dataset,
)
from .fields import MoabFieldInfo, PyneFieldInfo
from .io import IOHandlerMoabH5MHex8



import numpy as np

from yt.units.dimensions import current_mks
from yt.units.unit_object import Unit  # type: ignore
from yt.utilities.chemical_formulas import compute_mu
from yt.utilities.lib.misc_utilities import obtain_relative_velocity_vector

from .derived_field import ValidateParameter, ValidateSpatial
from .field_plugin_registry import register_field_plugin
from .vector_operations import (
    create_averaged_field,
    create_magnitude_field,
    create_vector_fields,
)


@register_field_plugin
def setup_fluid_fields(registry, ftype="gas", slice_info=None):
    pc = registry.ds.units.physical_constants
    # slice_info would be the left, the right, and the factor.
    # For example, with the old Enzo-ZEUS fields, this would be:
    # slice(None, -2, None)
    # slice(1, -1, None)
    # 1.0
    # Otherwise, we default to a centered difference.
    if slice_info is None:
        sl_left = slice(None, -2, None)
        sl_right = slice(2, None, None)
        div_fac = 2.0
    else:
        sl_left, sl_right, div_fac = slice_info

    unit_system = registry.ds.unit_system

    if unit_system.base_units[current_mks] is None:
        mag_units = "magnetic_field_cgs"
    else:
        mag_units = "magnetic_field_mks"

    create_vector_fields(
        registry, "velocity", unit_system["velocity"], ftype, slice_info
    )

    create_vector_fields(
        registry, "magnetic_field", unit_system[mag_units], ftype, slice_info
    )

    def _cell_mass(field, data):
        return data[ftype, "density"] * data[ftype, "cell_volume"]

    registry.add_field(
        (ftype, "cell_mass"),
        sampling_type="cell",
        function=_cell_mass,
        units=unit_system["mass"],
    )
    registry.alias((ftype, "mass"), (ftype, "cell_mass"))

    # momentum
    def momentum_xyz(v):
        def _momm(field, data):
            return data["gas", "mass"] * data["gas", f"velocity_{v}"]

        def _momd(field, data):
            return data["gas", "density"] * data["gas", f"velocity_{v}"]

        return _momm, _momd

    for v in registry.ds.coordinates.axis_order:
        _momm, _momd = momentum_xyz(v)
        registry.add_field(
            ("gas", f"momentum_{v}"),
            sampling_type="local",
            function=_momm,
            units=unit_system["momentum"],
        )
        registry.add_field(
            ("gas", f"momentum_density_{v}"),
            sampling_type="local",
            function=_momd,
            units=unit_system["density"] * unit_system["velocity"],
        )

    def _sound_speed(field, data):
        tr = data.ds.gamma * data[ftype, "pressure"] / data[ftype, "density"]
        return np.sqrt(tr)

    registry.add_field(
        (ftype, "sound_speed"),
        sampling_type="local",
        function=_sound_speed,
        units=unit_system["velocity"],
    )

    def _radial_mach_number(field, data):
        """Radial component of M{|v|/c_sound}"""
        tr = data[ftype, "radial_velocity"] / data[ftype, "sound_speed"]
        return np.abs(tr)

    registry.add_field(
        (ftype, "radial_mach_number"),
        sampling_type="local",
        function=_radial_mach_number,
        units="",
    )

    def _kinetic_energy_density(field, data):
        v = obtain_relative_velocity_vector(data)
        return 0.5 * data[ftype, "density"] * (v**2).sum(axis=0)

    registry.add_field(
        (ftype, "kinetic_energy_density"),
        sampling_type="local",
        function=_kinetic_energy_density,
        units=unit_system["pressure"],
        validators=[ValidateParameter("bulk_velocity")],
    )

    def _mach_number(field, data):
        """M{|v|/c_sound}"""
        return data[ftype, "velocity_magnitude"] / data[ftype, "sound_speed"]

    registry.add_field(
        (ftype, "mach_number"), sampling_type="local", function=_mach_number, units=""
    )

    def _courant_time_step(field, data):
        t1 = data[ftype, "dx"] / (
            data[ftype, "sound_speed"] + np.abs(data[ftype, "velocity_x"])
        )
        t2 = data[ftype, "dy"] / (
            data[ftype, "sound_speed"] + np.abs(data[ftype, "velocity_y"])
        )
        t3 = data[ftype, "dz"] / (
            data[ftype, "sound_speed"] + np.abs(data[ftype, "velocity_z"])
        )
        tr = np.minimum(np.minimum(t1, t2), t3)
        return tr

    registry.add_field(
        (ftype, "courant_time_step"),
        sampling_type="cell",
        function=_courant_time_step,
        units=unit_system["time"],
    )

    def _pressure(field, data):
        """M{(Gamma-1.0)*rho*E}"""
        tr = (data.ds.gamma - 1.0) * (
            data[ftype, "density"] * data[ftype, "specific_thermal_energy"]
        )
        return tr

    registry.add_field(
        (ftype, "pressure"),
        sampling_type="local",
        function=_pressure,
        units=unit_system["pressure"],
    )

    def _kT(field, data):
        return (pc.kboltz * data[ftype, "temperature"]).in_units("keV")

    registry.add_field(
        (ftype, "kT"),
        sampling_type="local",
        function=_kT,
        units="keV",
        display_name="Temperature",
    )

    def _metallicity(field, data):
        return data[ftype, "metal_density"] / data[ftype, "density"]

    registry.add_field(
        (ftype, "metallicity"),
        sampling_type="local",
        function=_metallicity,
        units="Zsun",
    )

    def _metal_mass(field, data):
        Z = data[ftype, "metallicity"].to("dimensionless")
        return Z * data[ftype, "mass"]

    registry.add_field(
        (ftype, "metal_mass"),
        sampling_type="local",
        function=_metal_mass,
        units=unit_system["mass"],
    )

    if len(registry.ds.field_info.species_names) > 0:

        def _number_density(field, data):
            field_data = np.zeros_like(
                data["gas", f"{data.ds.field_info.species_names[0]}_number_density"]
            )
            for species in data.ds.field_info.species_names:
                field_data += data["gas", f"{species}_number_density"]
            return field_data

    else:

        def _number_density(field, data):
            mu = getattr(data.ds, "mu", compute_mu(data.ds.default_species_fields))
            return data[ftype, "density"] / (pc.mh * mu)

    registry.add_field(
        (ftype, "number_density"),
        sampling_type="local",
        function=_number_density,
        units=unit_system["number_density"],
    )

    def _mean_molecular_weight(field, data):
        return data[ftype, "density"] / (pc.mh * data[ftype, "number_density"])

    registry.add_field(
        (ftype, "mean_molecular_weight"),
        sampling_type="local",
        function=_mean_molecular_weight,
        units="",
    )

    setup_gradient_fields(
        registry, (ftype, "pressure"), unit_system["pressure"], slice_info
    )

    setup_gradient_fields(
        registry, (ftype, "density"), unit_system["density"], slice_info
    )

    create_averaged_field(
        registry,
        "density",
        unit_system["density"],
        ftype=ftype,
        slice_info=slice_info,
        weight="mass",
    )


def setup_gradient_fields(registry, grad_field, field_units, slice_info=None):
    assert isinstance(grad_field, tuple)
    ftype, fname = grad_field
    if slice_info is None:
        sl_left = slice(None, -2, None)
        sl_right = slice(2, None, None)
        div_fac = 2.0
    else:
        sl_left, sl_right, div_fac = slice_info
    slice_3d = (slice(1, -1), slice(1, -1), slice(1, -1))

    def grad_func(axi, ax):
        slice_3dl = slice_3d[:axi] + (sl_left,) + slice_3d[axi + 1 :]
        slice_3dr = slice_3d[:axi] + (sl_right,) + slice_3d[axi + 1 :]

        def func(field, data):
            block_order = getattr(data, "_block_order", "C")
            if block_order == "F":
                # Fortran-ordering: we need to swap axes here and
                # reswap below
                field_data = data[grad_field].swapaxes(0, 2)
            else:
                field_data = data[grad_field]
            dx = div_fac * data[ftype, f"d{ax}"]
            if ax == "theta":
                dx *= data[ftype, "r"]
            if ax == "phi":
                dx *= data[ftype, "r"] * np.sin(data[ftype, "theta"])
            f = field_data[slice_3dr] / dx[slice_3d]
            f -= field_data[slice_3dl] / dx[slice_3d]
            new_field = np.zeros_like(data[grad_field], dtype=np.float64)
            new_field = data.ds.arr(new_field, field_data.units / dx.units)
            new_field[slice_3d] = f

            if block_order == "F":
                new_field = new_field.swapaxes(0, 2)

            return new_field

        return func

    field_units = Unit(field_units, registry=registry.ds.unit_registry)
    grad_units = field_units / registry.ds.unit_system["length"]

    for axi, ax in enumerate(registry.ds.coordinates.axis_order):
        f = grad_func(axi, ax)
        registry.add_field(
            (ftype, f"{fname}_gradient_{ax}"),
            sampling_type="local",
            function=f,
            validators=[ValidateSpatial(1, [grad_field])],
            units=grad_units,
        )

    create_magnitude_field(
        registry,
        f"{fname}_gradient",
        grad_units,
        ftype=ftype,
        validators=[ValidateSpatial(1, [grad_field])],
    )



import numpy as np
import numpy.testing as npt
import pytest

from yt.testing import fake_amr_ds
from yt.visualization.api import SlicePlot


def test_float_vmin_then_set_unit():
    field = ("gas", "density")
    ds = fake_amr_ds(fields=[field], units=["g/cm**3"])

    p = SlicePlot(ds, "x", field)
    p.set_buff_size(16)

    p.render()
    cb = p.plots[field].image.colorbar
    raw_lims = np.array((cb.vmin, cb.vmax))
    desired_lims = raw_lims.copy()
    desired_lims[0] = 1e-2

    p.set_zlim(field, zmin=desired_lims[0])

    p.render()
    cb = p.plots[field].image.colorbar
    new_lims = np.array((cb.vmin, cb.vmax))
    npt.assert_almost_equal(new_lims, desired_lims)

    # 1 g/cm**3 == 1000 kg/m**3
    p.set_unit(field, "kg/m**3")
    p.render()

    cb = p.plots[field].image.colorbar
    new_lims = np.array((cb.vmin, cb.vmax))
    npt.assert_almost_equal(new_lims, 1000 * desired_lims)


def test_set_unit_then_float_vmin():
    field = ("gas", "density")
    ds = fake_amr_ds(fields=[field], units=["g/cm**3"])

    p = SlicePlot(ds, "x", field)
    p.set_buff_size(16)

    p.set_unit(field, "kg/m**3")
    p.set_zlim(field, zmin=1)
    p.render()
    cb = p.plots[field].image.colorbar
    assert cb.vmin == 1.0


def test_reset_zlim():
    field = ("gas", "density")
    ds = fake_amr_ds(fields=[field], units=["g/cm**3"])

    p = SlicePlot(ds, "x", field)
    p.set_buff_size(16)

    p.render()
    cb = p.plots[field].image.colorbar
    raw_lims = np.array((cb.vmin, cb.vmax))

    # set a new zmin value
    delta = np.diff(raw_lims)[0]
    p.set_zlim(field, zmin=raw_lims[0] + delta / 2)

    # passing "min" should restore default limit
    p.set_zlim(field, zmin="min")
    p.render()

    cb = p.plots[field].image.colorbar
    new_lims = np.array((cb.vmin, cb.vmax))
    npt.assert_array_equal(new_lims, raw_lims)


def test_set_dynamic_range_with_vmin():
    field = ("gas", "density")
    ds = fake_amr_ds(fields=[field], units=["g/cm**3"])

    p = SlicePlot(ds, "x", field)
    p.set_buff_size(16)

    zmin = 1e-2
    p.set_zlim(field, zmin=zmin, dynamic_range=2)

    p.render()
    cb = p.plots[field].image.colorbar
    new_lims = np.array((cb.vmin, cb.vmax))
    npt.assert_almost_equal(new_lims, (zmin, 2 * zmin))


def test_set_dynamic_range_with_vmax():
    field = ("gas", "density")
    ds = fake_amr_ds(fields=[field], units=["g/cm**3"])

    p = SlicePlot(ds, "x", field)
    p.set_buff_size(16)

    zmax = 1
    p.set_zlim(field, zmax=zmax, dynamic_range=2)

    p.render()
    cb = p.plots[field].image.colorbar
    new_lims = np.array((cb.vmin, cb.vmax))
    npt.assert_almost_equal(new_lims, (zmax / 2, zmax))


def test_set_dynamic_range_with_min():
    field = ("gas", "density")
    ds = fake_amr_ds(fields=[field], units=["g/cm**3"])

    p = SlicePlot(ds, "x", field)
    p.set_buff_size(16)

    p.render()
    cb = p.plots[field].image.colorbar
    vmin = cb.vmin

    p.set_zlim(field, zmin="min", dynamic_range=2)

    p.render()
    cb = p.plots[field].image.colorbar
    new_lims = np.array((cb.vmin, cb.vmax))
    npt.assert_almost_equal(new_lims, (vmin, 2 * vmin))


def test_set_dynamic_range_with_None():
    field = ("gas", "density")
    ds = fake_amr_ds(fields=[field], units=["g/cm**3"])

    p = SlicePlot(ds, "x", field)
    p.set_buff_size(16)

    p.render()
    cb = p.plots[field].image.colorbar
    vmin = cb.vmin

    with pytest.deprecated_call(match="Passing `zmin=None` explicitly is deprecated"):
        p.set_zlim(field, zmin=None, dynamic_range=2)

    p.render()
    cb = p.plots[field].image.colorbar
    new_lims = np.array((cb.vmin, cb.vmax))
    npt.assert_almost_equal(new_lims, (vmin, 2 * vmin))



import random

import numpy as np
from numpy.testing import assert_equal, assert_raises

from yt.loaders import load_amr_grids


def setup_test_ds():
    """Prepare setup specific environment"""
    grid_data = [
        {
            "left_edge": [0.0, 0.0, 0.0],
            "right_edge": [1.0, 1.0, 1.0],
            "level": 0,
            "dimensions": [16, 16, 16],
        },
        {
            "left_edge": [0.25, 0.25, 0.25],
            "right_edge": [0.75, 0.75, 0.75],
            "level": 1,
            "dimensions": [16, 16, 16],
        },
        {
            "left_edge": [0.25, 0.25, 0.375],
            "right_edge": [0.5, 0.5, 0.625],
            "level": 2,
            "dimensions": [16, 16, 16],
        },
        {
            "left_edge": [0.5, 0.5, 0.375],
            "right_edge": [0.75, 0.75, 0.625],
            "level": 2,
            "dimensions": [16, 16, 16],
        },
        {
            "left_edge": [0.3125, 0.3125, 0.4375],
            "right_edge": [0.4375, 0.4375, 0.5625],
            "level": 3,
            "dimensions": [16, 16, 16],
        },
        {
            "left_edge": [0.5625, 0.5625, 0.4375],
            "right_edge": [0.6875, 0.6875, 0.5625],
            "level": 3,
            "dimensions": [16, 16, 16],
        },
    ]

    for grid in grid_data:
        grid["density"] = (
            np.random.random(grid["dimensions"]) * 2 ** grid["level"],
            "g/cm**3",
        )
    return load_amr_grids(grid_data, [16, 16, 16])


def test_grid_tree():
    """Main test suite for GridTree"""
    test_ds = setup_test_ds()
    grid_tree = test_ds.index._get_grid_tree()
    indices, levels, nchild, children = grid_tree.return_tree_info()

    grid_levels = [grid.Level for grid in test_ds.index.grids]

    grid_indices = [grid.id - grid._id_offset for grid in test_ds.index.grids]
    grid_nchild = [len(grid.Children) for grid in test_ds.index.grids]

    assert_equal(levels, grid_levels)
    assert_equal(indices, grid_indices)
    assert_equal(nchild, grid_nchild)

    for i, grid in enumerate(test_ds.index.grids):
        if grid_nchild[i] > 0:
            grid_children = np.array(
                [child.id - child._id_offset for child in grid.Children]
            )
            assert_equal(grid_children, children[i])


def test_find_points():
    """Main test suite for MatchPoints"""
    num_points = 100
    test_ds = setup_test_ds()
    randx = np.random.uniform(
        low=test_ds.domain_left_edge[0],
        high=test_ds.domain_right_edge[0],
        size=num_points,
    )
    randy = np.random.uniform(
        low=test_ds.domain_left_edge[1],
        high=test_ds.domain_right_edge[1],
        size=num_points,
    )
    randz = np.random.uniform(
        low=test_ds.domain_left_edge[2],
        high=test_ds.domain_right_edge[2],
        size=num_points,
    )

    point_grids, point_grid_inds = test_ds.index._find_points(randx, randy, randz)

    grid_inds = np.zeros((num_points), dtype="int64")

    for ind, ixx, iyy, izz in zip(range(num_points), randx, randy, randz, strict=True):
        pos = np.array([ixx, iyy, izz])
        pt_level = -1

        for grid in test_ds.index.grids:
            if (
                np.all(pos >= grid.LeftEdge)
                and np.all(pos <= grid.RightEdge)
                and grid.Level > pt_level
            ):
                pt_level = grid.Level
                grid_inds[ind] = grid.id - grid._id_offset

    assert_equal(point_grid_inds, grid_inds)

    # Test whether find_points works for lists
    point_grids, point_grid_inds = test_ds.index._find_points(
        randx.tolist(), randy.tolist(), randz.tolist()
    )
    assert_equal(point_grid_inds, grid_inds)

    # Test if find_points works for scalar
    ind = random.randint(0, num_points - 1)
    point_grids, point_grid_inds = test_ds.index._find_points(
        randx[ind], randy[ind], randz[ind]
    )
    assert_equal(point_grid_inds, grid_inds[ind])

    # Test if find_points fails properly for non equal indices' array sizes
    assert_raises(ValueError, test_ds.index._find_points, [0], 1.0, [2, 3])


def test_grid_arrays_view():
    ds = setup_test_ds()
    tree = ds.index._get_grid_tree()
    grid_arr = tree.grid_arrays
    assert_equal(grid_arr["left_edge"], ds.index.grid_left_edge)
    assert_equal(grid_arr["right_edge"], ds.index.grid_right_edge)
    assert_equal(grid_arr["dims"], ds.index.grid_dimensions)
    assert_equal(grid_arr["level"], ds.index.grid_levels[:, 0])



from itertools import groupby

import numpy as np

from yt.geometry.selection_routines import AlwaysSelector
from yt.utilities.io_handler import BaseIOHandler
from yt.utilities.logger import ytLogger as mylog

# -----------------------------------------------------------------------------
# GAMER shares a similar HDF5 format, and thus io.py as well, with FLASH
# -----------------------------------------------------------------------------


# group grids with consecutive indices together to improve the I/O performance
# --> grids are assumed to be sorted into ascending numerical order already
def grid_sequences(grids):
    for _k, g in groupby(enumerate(grids), lambda i_x: i_x[0] - i_x[1].id):
        seq = [v[1] for v in g]
        yield seq


def particle_sequences(grids):
    for _k, g in groupby(enumerate(grids), lambda i_x: i_x[0] - i_x[1].id):
        seq = [v[1] for v in g]
        yield seq[0], seq[-1]


class IOHandlerGAMER(BaseIOHandler):
    _particle_reader = False
    _dataset_type = "gamer"

    def __init__(self, ds):
        super().__init__(ds)
        self._handle = ds._handle
        self._group_grid = ds._group_grid
        self._group_particle = ds._group_particle
        self._field_dtype = "float64"  # fixed even when FLOAT8 is off
        self._particle_handle = ds._particle_handle
        self.patch_size = ds.parameters["PatchSize"] * ds.refine_by
        self.pgroup = ds.refine_by**3  # number of patches in a patch group

    def _read_particle_coords(self, chunks, ptf):
        chunks = list(chunks)  # generator --> list
        p_idx = self.ds.index._particle_indices

        # shortcuts
        par_posx = self._group_particle["ParPosX"]
        par_posy = self._group_particle["ParPosY"]
        par_posz = self._group_particle["ParPosZ"]

        # currently GAMER does not support multiple particle types
        assert len(ptf) == 1
        ptype = list(ptf.keys())[0]

        for chunk in chunks:
            for g1, g2 in particle_sequences(chunk.objs):
                start = p_idx[g1.id]
                end = p_idx[g2.id + 1]
                x = np.asarray(par_posx[start:end], dtype=self._field_dtype)
                y = np.asarray(par_posy[start:end], dtype=self._field_dtype)
                z = np.asarray(par_posz[start:end], dtype=self._field_dtype)
                yield ptype, (x, y, z), 0.0

    def _read_particle_fields(self, chunks, ptf, selector):
        chunks = list(chunks)  # generator --> list
        p_idx = self.ds.index._particle_indices

        # shortcuts
        par_posx = self._group_particle["ParPosX"]
        par_posy = self._group_particle["ParPosY"]
        par_posz = self._group_particle["ParPosZ"]

        # currently GAMER does not support multiple particle types
        assert len(ptf) == 1
        ptype = list(ptf.keys())[0]
        pfields = ptf[ptype]

        for chunk in chunks:
            for g1, g2 in particle_sequences(chunk.objs):
                start = p_idx[g1.id]
                end = p_idx[g2.id + 1]
                x = np.asarray(par_posx[start:end], dtype=self._field_dtype)
                y = np.asarray(par_posy[start:end], dtype=self._field_dtype)
                z = np.asarray(par_posz[start:end], dtype=self._field_dtype)

                mask = selector.select_points(x, y, z, 0.0)
                if mask is None:
                    continue

                for field in pfields:
                    data = self._group_particle[field][start:end]
                    yield (ptype, field), data[mask]

    def _read_fluid_selection(self, chunks, selector, fields, size):
        chunks = list(chunks)  # generator --> list

        if any((ftype != "gamer" for ftype, fname in fields)):
            raise NotImplementedError

        rv = {}
        for field in fields:
            rv[field] = np.empty(size, dtype=self._field_dtype)

        ng = sum(len(c.objs) for c in chunks)  # c.objs is a list of grids
        mylog.debug(
            "Reading %s cells of %s fields in %s grids",
            size,
            [f2 for f1, f2 in fields],
            ng,
        )

        # shortcuts
        ps2 = self.patch_size
        ps1 = ps2 // 2

        for field in fields:
            ds = self._group_grid[field[1]]
            offset = 0
            for chunk in chunks:
                for gs in grid_sequences(chunk.objs):
                    start = (gs[0].id) * self.pgroup
                    end = (gs[-1].id + 1) * self.pgroup
                    buf = ds[start:end, :, :, :]
                    ngrid = len(gs)
                    data = np.empty((ngrid, ps2, ps2, ps2), dtype=self._field_dtype)

                    for g in range(ngrid):
                        pid0 = g * self.pgroup
                        data[g, 0:ps1, 0:ps1, 0:ps1] = buf[pid0 + 0, :, :, :]
                        data[g, 0:ps1, 0:ps1, ps1:ps2] = buf[pid0 + 1, :, :, :]
                        data[g, 0:ps1, ps1:ps2, 0:ps1] = buf[pid0 + 2, :, :, :]
                        data[g, ps1:ps2, 0:ps1, 0:ps1] = buf[pid0 + 3, :, :, :]
                        data[g, 0:ps1, ps1:ps2, ps1:ps2] = buf[pid0 + 4, :, :, :]
                        data[g, ps1:ps2, ps1:ps2, 0:ps1] = buf[pid0 + 5, :, :, :]
                        data[g, ps1:ps2, 0:ps1, ps1:ps2] = buf[pid0 + 6, :, :, :]
                        data[g, ps1:ps2, ps1:ps2, ps1:ps2] = buf[pid0 + 7, :, :, :]

                    data = data.transpose()

                    for i, g in enumerate(gs):
                        offset += g.select(selector, data[..., i], rv[field], offset)
        return rv

    def _read_chunk_data(self, chunk, fields):
        rv = {}
        if len(chunk.objs) == 0:
            return rv

        for g in chunk.objs:
            rv[g.id] = {}

        # Split into particles and non-particles
        fluid_fields, particle_fields = [], []
        for ftype, fname in fields:
            if ftype in self.ds.particle_types:
                particle_fields.append((ftype, fname))
            else:
                fluid_fields.append((ftype, fname))

        # particles
        if len(particle_fields) > 0:
            selector = AlwaysSelector(self.ds)
            rv.update(self._read_particle_selection([chunk], selector, particle_fields))

        # fluid
        if len(fluid_fields) == 0:
            return rv

        ps2 = self.patch_size
        ps1 = ps2 // 2

        for field in fluid_fields:
            ds = self._group_grid[field[1]]

            for gs in grid_sequences(chunk.objs):
                start = (gs[0].id) * self.pgroup
                end = (gs[-1].id + 1) * self.pgroup
                buf = ds[start:end, :, :, :]
                ngrid = len(gs)
                data = np.empty((ngrid, ps2, ps2, ps2), dtype=self._field_dtype)

                for g in range(ngrid):
                    pid0 = g * self.pgroup
                    data[g, 0:ps1, 0:ps1, 0:ps1] = buf[pid0 + 0, :, :, :]
                    data[g, 0:ps1, 0:ps1, ps1:ps2] = buf[pid0 + 1, :, :, :]
                    data[g, 0:ps1, ps1:ps2, 0:ps1] = buf[pid0 + 2, :, :, :]
                    data[g, ps1:ps2, 0:ps1, 0:ps1] = buf[pid0 + 3, :, :, :]
                    data[g, 0:ps1, ps1:ps2, ps1:ps2] = buf[pid0 + 4, :, :, :]
                    data[g, ps1:ps2, ps1:ps2, 0:ps1] = buf[pid0 + 5, :, :, :]
                    data[g, ps1:ps2, 0:ps1, ps1:ps2] = buf[pid0 + 6, :, :, :]
                    data[g, ps1:ps2, ps1:ps2, ps1:ps2] = buf[pid0 + 7, :, :, :]

                data = data.transpose()

                for i, g in enumerate(gs):
                    rv[g.id][field] = data[..., i]
        return rv



import os

from yt.utilities.configure import YTConfig, config_dir, configuration_callbacks

ytcfg_defaults = {}

ytcfg_defaults["yt"] = {
    "serialize": False,
    "only_deserialize": False,
    "time_functions": False,
    "colored_logs": False,
    "suppress_stream_logging": False,
    "stdout_stream_logging": False,
    "log_level": 20,
    "inline": False,
    "num_threads": -1,
    "store_parameter_files": False,
    "parameter_file_store": "parameter_files.csv",
    "maximum_stored_datasets": 500,
    "skip_dataset_cache": True,
    "load_field_plugins": False,
    "plugin_filename": "my_plugins.py",
    "parallel_traceback": False,
    "pasteboard_repo": "",
    "reconstruct_index": True,
    "test_storage_dir": "/does/not/exist",
    "test_data_dir": "/does/not/exist",
    "enzo_db": "",
    "answer_testing_tolerance": 3,
    "answer_testing_bitwise": False,
    "gold_standard_filename": "gold311",
    "local_standard_filename": "local001",
    "answer_tests_url": "http://answers.yt-project.org/{1}_{2}",
    "sketchfab_api_key": "None",
    "imagebin_api_key": "e1977d9195fe39e",
    "imagebin_upload_url": "https://api.imgur.com/3/image",
    "imagebin_delete_url": "https://api.imgur.com/3/image/{delete_hash}",
    "curldrop_upload_url": "http://use.yt/upload",
    "thread_field_detection": False,
    "ignore_invalid_unit_operation_errors": False,
    "chunk_size": 1000,
    "xray_data_dir": "/does/not/exist",
    "supp_data_dir": "/does/not/exist",
    "default_colormap": "cmyt.arbre",
    "ray_tracing_engine": "yt",
    "internals": {
        "within_testing": False,
        "parallel": False,
        "strict_requires": False,
        "global_parallel_rank": 0,
        "global_parallel_size": 1,
        "topcomm_parallel_rank": 0,
        "topcomm_parallel_size": 1,
        "command_line": False,
    },
}


# For backward compatibility, do not use these vars internally in yt
CONFIG_DIR = config_dir()


_global_config_file = YTConfig.get_global_config_file()
_local_config_file = YTConfig.get_local_config_file()

# Load the config
ytcfg = YTConfig()
ytcfg.update(ytcfg_defaults, metadata={"source": "defaults"})

# Try loading the local config first, otherwise fall back to global config
if os.path.exists(_local_config_file):
    ytcfg.read(_local_config_file)
elif os.path.exists(_global_config_file):
    ytcfg.read(_global_config_file)


def _setup_postinit_configuration():
    """This is meant to be run last in yt.__init__"""
    for callback in configuration_callbacks:
        callback(ytcfg)



import numpy as np

import yt.utilities.lib.interpolators as lib
from yt.funcs import mylog


class UnilinearFieldInterpolator:
    def __init__(self, table, boundaries, field_names, truncate=False):
        r"""Initialize a 1D interpolator for field data.

        table : array
            The data table over which interpolation is performed.
        boundaries: tuple or array
            If a tuple, this should specify the upper and lower bounds
            for the bins of the data table.  This assumes the bins are
            evenly spaced.  If an array, this specifies the bins
            explicitly.
        field_names: str
            Name of the field to be used as input data for interpolation.
        truncate : bool
            If False, an exception is raised if the input values are
            outside the bounds of the table.  If True, extrapolation is
            performed.

        Examples
        --------

        ad = ds.all_data()
        table_data = np.random.random(64)
        interp = UnilinearFieldInterpolator(table_data, (0.0, 1.0), "x",
                                            truncate=True)
        field_data = interp(ad)

        """
        self.table = table.astype("float64")
        self.truncate = truncate
        self.x_name = field_names
        if isinstance(boundaries, np.ndarray):
            if boundaries.size != table.shape[0]:
                mylog.error("Bins array not the same length as the data.")
                raise ValueError
            self.x_bins = boundaries
        else:
            x0, x1 = boundaries
            self.x_bins = np.linspace(x0, x1, table.shape[0], dtype="float64")

    def __call__(self, data_object):
        orig_shape = data_object[self.x_name].shape
        x_vals = data_object[self.x_name].ravel().astype("float64")

        x_i = (np.digitize(x_vals, self.x_bins) - 1).astype("int32")
        if np.any((x_i == -1) | (x_i == len(self.x_bins) - 1)):
            if not self.truncate:
                mylog.error(
                    "Sorry, but your values are outside "
                    "the table!  Dunno what to do, so dying."
                )
                mylog.error("Error was in: %s", data_object)
                raise ValueError
            else:
                x_i = np.minimum(np.maximum(x_i, 0), len(self.x_bins) - 2)

        my_vals = np.zeros(x_vals.shape, dtype="float64")
        lib.UnilinearlyInterpolate(self.table, x_vals, self.x_bins, x_i, my_vals)
        return my_vals.reshape(orig_shape)


class BilinearFieldInterpolator:
    def __init__(self, table, boundaries, field_names, truncate=False):
        r"""Initialize a 2D interpolator for field data.

        table : array
            The data table over which interpolation is performed.
        boundaries: tuple
            Either a tuple of lower and upper bounds for the x and y bins
            given as (x0, x1, y0, y1) or a tuple of two arrays containing the
            x and y bins.
        field_names: list
            Names of the fields to be used as input data for interpolation.
        truncate : bool
            If False, an exception is raised if the input values are
            outside the bounds of the table.  If True, extrapolation is
            performed.

        Examples
        --------

        ad = ds.all_data()
        table_data = np.random.random((64, 64))
        interp = BilinearFieldInterpolator(table_data, (0.0, 1.0, 0.0, 1.0),
                                           ["x", "y"],
                                           truncate=True)
        field_data = interp(ad)

        """
        self.table = table.astype("float64")
        self.truncate = truncate
        self.x_name, self.y_name = field_names
        if len(boundaries) == 4:
            x0, x1, y0, y1 = boundaries
            self.x_bins = np.linspace(x0, x1, table.shape[0], dtype="float64")
            self.y_bins = np.linspace(y0, y1, table.shape[1], dtype="float64")
        elif len(boundaries) == 2:
            if boundaries[0].size != table.shape[0]:
                mylog.error("X bins array not the same length as the data.")
                raise ValueError
            if boundaries[1].size != table.shape[1]:
                mylog.error("Y bins array not the same length as the data.")
                raise ValueError
            self.x_bins = boundaries[0]
            self.y_bins = boundaries[1]
        else:
            mylog.error(
                "Boundaries must be given as (x0, x1, y0, y1) or as (x_bins, y_bins)"
            )
            raise ValueError

    def __call__(self, data_object):
        orig_shape = data_object[self.x_name].shape
        x_vals = data_object[self.x_name].ravel().astype("float64")
        y_vals = data_object[self.y_name].ravel().astype("float64")

        x_i = (np.digitize(x_vals, self.x_bins) - 1).astype("int32")
        y_i = (np.digitize(y_vals, self.y_bins) - 1).astype("int32")
        if np.any((x_i == -1) | (x_i == len(self.x_bins) - 1)) or np.any(
            (y_i == -1) | (y_i == len(self.y_bins) - 1)
        ):
            if not self.truncate:
                mylog.error(
                    "Sorry, but your values are outside "
                    "the table!  Dunno what to do, so dying."
                )
                mylog.error("Error was in: %s", data_object)
                raise ValueError
            else:
                x_i = np.minimum(np.maximum(x_i, 0), len(self.x_bins) - 2)
                y_i = np.minimum(np.maximum(y_i, 0), len(self.y_bins) - 2)

        my_vals = np.zeros(x_vals.shape, dtype="float64")
        lib.BilinearlyInterpolate(
            self.table, x_vals, y_vals, self.x_bins, self.y_bins, x_i, y_i, my_vals
        )
        return my_vals.reshape(orig_shape)


class TrilinearFieldInterpolator:
    def __init__(self, table, boundaries, field_names, truncate=False):
        r"""Initialize a 3D interpolator for field data.

        table : array
            The data table over which interpolation is performed.
        boundaries: tuple
            Either a tuple of lower and upper bounds for the x, y, and z bins
            given as (x0, x1, y0, y1, z0, z1) or a tuple of three arrays
            containing the x, y, and z bins.
        field_names: list
            Names of the fields to be used as input data for interpolation.
        truncate : bool
            If False, an exception is raised if the input values are
            outside the bounds of the table.  If True, extrapolation is
            performed.

        Examples
        --------

        ad = ds.all_data()
        table_data = np.random.random((64, 64, 64))
        interp = TrilinearFieldInterpolator(table_data,
                                            (0.0, 1.0, 0.0, 1.0, 0.0, 1.0),
                                            ["x", "y", "z"],
                                            truncate=True)
        field_data = interp(ad)

        """
        self.table = table.astype("float64")
        self.truncate = truncate
        self.x_name, self.y_name, self.z_name = field_names
        if len(boundaries) == 6:
            x0, x1, y0, y1, z0, z1 = boundaries
            self.x_bins = np.linspace(x0, x1, table.shape[0], dtype="float64")
            self.y_bins = np.linspace(y0, y1, table.shape[1], dtype="float64")
            self.z_bins = np.linspace(z0, z1, table.shape[2], dtype="float64")
        elif len(boundaries) == 3:
            if boundaries[0].size != table.shape[0]:
                mylog.error("X bins array not the same length as the data.")
                raise ValueError
            if boundaries[1].size != table.shape[1]:
                mylog.error("Y bins array not the same length as the data.")
                raise ValueError
            if boundaries[2].size != table.shape[2]:
                mylog.error("Z bins array not the same length as the data.")
                raise ValueError
            self.x_bins = boundaries[0]
            self.y_bins = boundaries[1]
            self.z_bins = boundaries[2]
        else:
            mylog.error(
                "Boundaries must be given as (x0, x1, y0, y1, z0, z1) "
                "or as (x_bins, y_bins, z_bins)"
            )
            raise ValueError

    def __call__(self, data_object):
        orig_shape = data_object[self.x_name].shape
        x_vals = data_object[self.x_name].ravel().astype("float64")
        y_vals = data_object[self.y_name].ravel().astype("float64")
        z_vals = data_object[self.z_name].ravel().astype("float64")

        x_i = np.digitize(x_vals, self.x_bins).astype("int64") - 1
        y_i = np.digitize(y_vals, self.y_bins).astype("int64") - 1
        z_i = np.digitize(z_vals, self.z_bins).astype("int64") - 1
        if (
            np.any((x_i == -1) | (x_i == len(self.x_bins) - 1))
            or np.any((y_i == -1) | (y_i == len(self.y_bins) - 1))
            or np.any((z_i == -1) | (z_i == len(self.z_bins) - 1))
        ):
            if not self.truncate:
                mylog.error(
                    "Sorry, but your values are outside "
                    "the table!  Dunno what to do, so dying."
                )
                mylog.error("Error was in: %s", data_object)
                raise ValueError
            else:
                x_i = np.minimum(np.maximum(x_i, 0), len(self.x_bins) - 2)
                y_i = np.minimum(np.maximum(y_i, 0), len(self.y_bins) - 2)
                z_i = np.minimum(np.maximum(z_i, 0), len(self.z_bins) - 2)

        my_vals = np.zeros(x_vals.shape, dtype="float64")
        lib.TrilinearlyInterpolate(
            self.table,
            x_vals,
            y_vals,
            z_vals,
            self.x_bins,
            self.y_bins,
            self.z_bins,
            x_i,
            y_i,
            z_i,
            my_vals,
        )
        return my_vals.reshape(orig_shape)


class QuadrilinearFieldInterpolator:
    def __init__(self, table, boundaries, field_names, truncate=False):
        r"""Initialize a 4D interpolator for field data.

        table : array
            The data table over which interpolation is performed.
        boundaries: tuple
            Either a tuple of lower and upper bounds for the x, y, z, and w bins
            given as (x0, x1, y0, y1, z0, z1, w0, w1) or a tuple of four arrays
            containing the x, y, z, and w bins.
        field_names: list
            Names of the fields to be used as input data for interpolation.
        truncate : bool
            If False, an exception is raised if the input values are
            outside the bounds of the table.  If True, extrapolation is
            performed.

        Examples
        --------
        ad = ds.all_data()
        table_data = np.random.random((64, 64, 64, 64))
        interp = BilinearFieldInterpolator(table_data,
                                           (0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0),
                                           ["x", "y", "z", "w"],
                                           truncate=True)
        field_data = interp(ad)

        """
        self.table = table.astype("float64")
        self.truncate = truncate
        self.x_name, self.y_name, self.z_name, self.w_name = field_names
        if len(boundaries) == 8:
            x0, x1, y0, y1, z0, z1, w0, w1 = boundaries
            self.x_bins = np.linspace(x0, x1, table.shape[0]).astype("float64")
            self.y_bins = np.linspace(y0, y1, table.shape[1]).astype("float64")
            self.z_bins = np.linspace(z0, z1, table.shape[2]).astype("float64")
            self.w_bins = np.linspace(w0, w1, table.shape[3]).astype("float64")
        elif len(boundaries) == 4:
            if boundaries[0].size != table.shape[0]:
                mylog.error("X bins array not the same length as the data.")
                raise ValueError
            if boundaries[1].size != table.shape[1]:
                mylog.error("Y bins array not the same length as the data.")
                raise ValueError
            if boundaries[2].size != table.shape[2]:
                mylog.error("Z bins array not the same length as the data.")
                raise ValueError
            if boundaries[3].size != table.shape[3]:
                mylog.error("W bins array not the same length as the data.")
                raise ValueError
            self.x_bins = boundaries[0]
            self.y_bins = boundaries[1]
            self.z_bins = boundaries[2]
            self.w_bins = boundaries[3]
        else:
            mylog.error(
                "Boundaries must be given as (x0, x1, y0, y1, z0, z1, w0, w1) "
                "or as (x_bins, y_bins, z_bins, w_bins)"
            )
            raise ValueError

    def __call__(self, data_object):
        orig_shape = data_object[self.x_name].shape
        x_vals = data_object[self.x_name].ravel().astype("float64")
        y_vals = data_object[self.y_name].ravel().astype("float64")
        z_vals = data_object[self.z_name].ravel().astype("float64")
        w_vals = data_object[self.w_name].ravel().astype("float64")

        x_i = np.digitize(x_vals, self.x_bins).astype("int64") - 1
        y_i = np.digitize(y_vals, self.y_bins).astype("int64") - 1
        z_i = np.digitize(z_vals, self.z_bins).astype("int64") - 1
        w_i = np.digitize(w_vals, self.w_bins).astype("int64") - 1
        if (
            np.any((x_i == -1) | (x_i == len(self.x_bins) - 1))
            or np.any((y_i == -1) | (y_i == len(self.y_bins) - 1))
            or np.any((z_i == -1) | (z_i == len(self.z_bins) - 1))
            or np.any((w_i == -1) | (w_i == len(self.w_bins) - 1))
        ):
            if not self.truncate:
                mylog.error(
                    "Sorry, but your values are outside "
                    "the table!  Dunno what to do, so dying."
                )
                mylog.error("Error was in: %s", data_object)
                raise ValueError
            else:
                x_i = np.minimum(np.maximum(x_i, 0), len(self.x_bins) - 2)
                y_i = np.minimum(np.maximum(y_i, 0), len(self.y_bins) - 2)
                z_i = np.minimum(np.maximum(z_i, 0), len(self.z_bins) - 2)
                w_i = np.minimum(np.maximum(w_i, 0), len(self.w_bins) - 2)

        my_vals = np.zeros(x_vals.shape, dtype="float64")
        lib.QuadrilinearlyInterpolate(
            self.table,
            x_vals,
            y_vals,
            z_vals,
            w_vals,
            self.x_bins,
            self.y_bins,
            self.z_bins,
            self.w_bins,
            x_i,
            y_i,
            z_i,
            w_i,
            my_vals,
        )
        return my_vals.reshape(orig_shape)


def get_centers(ds, filename, center_cols, radius_col, unit="1"):
    """
    Return an iterator over EnzoSphere objects generated from the appropriate
    columns in *filename*.  Optionally specify the *unit* radius is in.
    """
    for line in open(filename):
        if line.startswith("#"):
            continue
        vals = line.split()
        x, y, z = (float(vals[i]) for i in center_cols)
        r = float(vals[radius_col])
        yield ds.sphere([x, y, z], r / ds[unit])



cimport cython
from libc.math cimport sqrt


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline cython.floating dot(const cython.floating[3] a,
                                const cython.floating[3] b) noexcept nogil:
    return a[0]*b[0] + a[1]*b[1] + a[2]*b[2]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline void cross(const cython.floating[3] a,
                       const cython.floating[3] b,
                       cython.floating c[3]) noexcept nogil:
    c[0] = a[1]*b[2] - a[2]*b[1]
    c[1] = a[2]*b[0] - a[0]*b[2]
    c[2] = a[0]*b[1] - a[1]*b[0]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline void subtract(const cython.floating[3] a,
                          const cython.floating[3] b,
                          cython.floating c[3]) noexcept nogil:
    c[0] = a[0] - b[0]
    c[1] = a[1] - b[1]
    c[2] = a[2] - b[2]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline cython.floating distance(const cython.floating[3] a,
                                     const cython.floating[3] b) noexcept nogil:
    return sqrt((a[0] - b[0])**2 + (a[1] - b[1])**2 +(a[2] - b[2])**2)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline void fma(const cython.floating f,
                     const cython.floating[3] a,
                     const cython.floating[3] b,
                     cython.floating[3] c) noexcept nogil:
    c[0] = f * a[0] + b[0]
    c[1] = f * a[1] + b[1]
    c[2] = f * a[2] + b[2]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline cython.floating L2_norm(const cython.floating[3] a) noexcept nogil:
    return sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2])



"""Module for cookbook testing


This test should be run from main yt directory.

Example:

      $ sed -e '/where/d' -i nose.cfg setup.cfg
      $ nosetests doc/source/cookbook/tests/test_cookbook.py -P -v
"""

import subprocess
import sys

from pathlib import Path


BLACKLIST = [
    "matplotlib-animation.py",
]


def test_recipe():
    """Dummy test grabbing all cookbook's recipes"""
    COOKBOOK_DIR = Path("doc", "source", "cookbook")
    for fname in sorted(COOKBOOK_DIR.glob("*.py")):
        if fname.name in BLACKLIST:
            continue
        check_recipe.description = f"Testing recipe: {fname.name}"
        yield check_recipe, ["python", str(fname)]


def check_recipe(cmd):
    """Run single recipe"""
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = proc.communicate()
    if out:
        sys.stdout.write(out.decode("utf8"))
    if err:
        sys.stderr.write(err.decode("utf8"))

    if proc.returncode != 0:
        retstderr = " ".join(cmd)
        retstderr += "\n\nTHIS IS THE REAL CAUSE OF THE FAILURE:\n"
        retstderr += err.decode("UTF-8") + "\n"
        raise subprocess.CalledProcessError(proc.returncode, retstderr)



cdef class PointSelector(SelectorObject):
    cdef public np.float64_t p[3]

    def __init__(self, dobj):
        cdef const np.float64_t[:] DLE = dobj.ds.domain_left_edge
        cdef const np.float64_t[:] DRE = dobj.ds.domain_right_edge
        for i in range(3):
            self.p[i] = _ensure_code(dobj.p[i])

            # ensure the point lies in the domain
            if self.periodicity[i]:
                self.p[i] = np.fmod(self.p[i], self.domain_width[i])
                if self.p[i] < DLE[i]:
                    self.p[i] += self.domain_width[i]
                elif self.p[i] >= DRE[i]:
                    self.p[i] -= self.domain_width[i]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) noexcept nogil:
        if (pos[0] - 0.5*dds[0] <= self.p[0] < pos[0]+0.5*dds[0] and
            pos[1] - 0.5*dds[1] <= self.p[1] < pos[1]+0.5*dds[1] and
            pos[2] - 0.5*dds[2] <= self.p[2] < pos[2]+0.5*dds[2]):
            return 1
        else:
            return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) noexcept nogil:
        cdef int i
        cdef np.float64_t dist, dist2 = 0
        for i in range(3):
            dist = self.periodic_difference(pos[i], self.p[i], i)
            dist2 += dist*dist
        if dist2 <= radius*radius: return 1
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil:
        # point definitely can only be in one cell
        if (left_edge[0] <= self.p[0] < right_edge[0] and
            left_edge[1] <= self.p[1] < right_edge[1] and
            left_edge[2] <= self.p[2] < right_edge[2]):
            return 1
        else:
            return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil:
        # point definitely can only be in one cell
        # Return 2 in all cases to indicate that the point only overlaps
        # portion of box
        if (left_edge[0] <= self.p[0] <= right_edge[0] and
            left_edge[1] <= self.p[1] <= right_edge[1] and
            left_edge[2] <= self.p[2] <= right_edge[2]):
            return 2
        else:
            return 0

    def _hash_vals(self):
        return (("p[0]", self.p[0]),
                ("p[1]", self.p[1]),
                ("p[2]", self.p[2]))

    def _get_state_attnames(self):
        return ('p', )

point_selector = PointSelector



import yt

# Load the dataset.
ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

# You can create a derived field by manipulating any existing derived fields
# in any way you choose.  In this case, let's just make a simple one:
# thermal_energy_density = 3/2 nkT


# First create a function which yields your new derived field
def thermal_energy_dens(field, data):
    return (3 / 2) * data["gas", "number_density"] * data["gas", "kT"]


# Then add it to your dataset and define the units
ds.add_field(
    ("gas", "thermal_energy_density"),
    units="erg/cm**3",
    function=thermal_energy_dens,
    sampling_type="cell",
)

# It will now show up in your derived_field_list
for i in sorted(ds.derived_field_list):
    print(i)

# Let's use it to make a projection
ad = ds.all_data()
yt.ProjectionPlot(
    ds,
    "x",
    ("gas", "thermal_energy_density"),
    weight_field=("gas", "density"),
    width=(200, "kpc"),
).save()



import yt

# load the dataset
ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

# create our plot
p = yt.ParticlePlot(
    ds,
    ("all", "particle_position_x"),
    ("all", "particle_position_y"),
    ("all", "particle_mass"),
    width=(0.5, 0.5),
)

# pick some appropriate units
p.set_axes_unit("kpc")
p.set_unit(("all", "particle_mass"), "Msun")

# save result
p.save()



import yt

# Create a time-series object.
sim = yt.load_simulation("enzo_tiny_cosmology/32Mpc_32.enzo", "Enzo")
sim.get_time_series(redshifts=[5, 4, 3, 2, 1, 0])

# Lists to hold profiles, labels, and plot specifications.
profiles = []
labels = []
plot_specs = []

# Loop over each dataset in the time-series.
for ds in sim:
    # Create a data container to hold the whole dataset.
    ad = ds.all_data()
    # Create a 1d profile of density vs. temperature.
    profiles.append(
        yt.create_profile(ad, [("gas", "density")], fields=[("gas", "temperature")])
    )
    # Add labels and linestyles.
    labels.append(f"z = {ds.current_redshift:.2f}")
    plot_specs.append(dict(linewidth=2, alpha=0.7))

# Create the profile plot from the list of profiles.
plot = yt.ProfilePlot.from_profiles(profiles, labels=labels, plot_specs=plot_specs)
# Save the image.
plot.save()



import numpy as np

from yt.data_objects.selection_objects.data_selection_objects import (
    YTSelectionContainer3D,
)
from yt.data_objects.static_output import Dataset
from yt.utilities.lib import bounding_volume_hierarchy
from yt.utilities.lib.image_samplers import (
    InterpolatedProjectionSampler,
    ProjectionSampler,
    VolumeRenderSampler,
)
from yt.utilities.on_demand_imports import NotAModule

try:
    from yt.utilities.lib.embree_mesh import mesh_traversal  # type: ignore
# Catch ValueError in case size of objects in Cython change
except (ImportError, ValueError):
    mesh_traversal = NotAModule("pyembree")


def data_source_or_all(data_source):
    if isinstance(data_source, Dataset):
        data_source = data_source.all_data()
    if not isinstance(data_source, (YTSelectionContainer3D, type(None))):
        raise RuntimeError(
            "The data_source is not a valid 3D data container.\n"
            "Expected an object of type YTSelectionContainer3D but received "
            f"an object of type {type(data_source)}."
        )
    return data_source


def new_mesh_sampler(camera, render_source, engine):
    params = ensure_code_unit_params(camera._get_sampler_params(render_source))
    args = (
        np.atleast_3d(params["vp_pos"]),
        np.atleast_3d(params["vp_dir"]),
        params["center"],
        params["bounds"],
        np.atleast_3d(params["image"]).astype("float64"),
        params["x_vec"],
        params["y_vec"],
        params["width"],
        render_source.volume_method,
    )
    kwargs = {"lens_type": params["lens_type"]}
    if engine == "embree":
        sampler = mesh_traversal.EmbreeMeshSampler(*args, **kwargs)
    elif engine == "yt":
        sampler = bounding_volume_hierarchy.BVHMeshSampler(*args, **kwargs)
    return sampler


def new_volume_render_sampler(camera, render_source):
    params = ensure_code_unit_params(camera._get_sampler_params(render_source))
    params.update(transfer_function=render_source.transfer_function)
    params.update(transfer_function=render_source.transfer_function)
    params.update(num_samples=render_source.num_samples)
    args = (
        np.atleast_3d(params["vp_pos"]),
        np.atleast_3d(params["vp_dir"]),
        params["center"],
        params["bounds"],
        params["image"],
        params["x_vec"],
        params["y_vec"],
        params["width"],
        render_source.volume_method,
        params["transfer_function"],
        params["num_samples"],
    )
    kwargs = {
        "lens_type": params["lens_type"],
    }
    if "camera_data" in params:
        kwargs["camera_data"] = params["camera_data"]
    if render_source.zbuffer is not None:
        kwargs["zbuffer"] = render_source.zbuffer.z
        args[4][:] = np.reshape(
            render_source.zbuffer.rgba[:],
            (camera.resolution[0], camera.resolution[1], 4),
        )
    else:
        kwargs["zbuffer"] = np.ones(params["image"].shape[:2], "float64")
    sampler = VolumeRenderSampler(*args, **kwargs)
    return sampler


def new_interpolated_projection_sampler(camera, render_source):
    params = ensure_code_unit_params(camera._get_sampler_params(render_source))
    params.update(transfer_function=render_source.transfer_function)
    params.update(num_samples=render_source.num_samples)
    args = (
        np.atleast_3d(params["vp_pos"]),
        np.atleast_3d(params["vp_dir"]),
        params["center"],
        params["bounds"],
        params["image"],
        params["x_vec"],
        params["y_vec"],
        params["width"],
        render_source.volume_method,
        params["num_samples"],
    )
    kwargs = {"lens_type": params["lens_type"]}
    if render_source.zbuffer is not None:
        kwargs["zbuffer"] = render_source.zbuffer.z
    else:
        kwargs["zbuffer"] = np.ones(params["image"].shape[:2], "float64")
    sampler = InterpolatedProjectionSampler(*args, **kwargs)
    return sampler


def new_projection_sampler(camera, render_source):
    params = ensure_code_unit_params(camera._get_sampler_params(render_source))
    params.update(transfer_function=render_source.transfer_function)
    params.update(num_samples=render_source.num_samples)
    args = (
        np.atleast_3d(params["vp_pos"]),
        np.atleast_3d(params["vp_dir"]),
        params["center"],
        params["bounds"],
        params["image"],
        params["x_vec"],
        params["y_vec"],
        params["width"],
        render_source.volume_method,
        params["num_samples"],
    )
    kwargs = {
        "lens_type": params["lens_type"],
    }
    if render_source.zbuffer is not None:
        kwargs["zbuffer"] = render_source.zbuffer.z
    else:
        kwargs["zbuffer"] = np.ones(params["image"].shape[:2], "float64")
    sampler = ProjectionSampler(*args, **kwargs)
    return sampler


def get_corners(le, re):
    return np.array(
        [
            [le[0], le[1], le[2]],
            [re[0], le[1], le[2]],
            [re[0], re[1], le[2]],
            [le[0], re[1], le[2]],
            [le[0], le[1], re[2]],
            [re[0], le[1], re[2]],
            [re[0], re[1], re[2]],
            [le[0], re[1], re[2]],
        ],
        dtype="float64",
    )


def ensure_code_unit_params(params):
    for param_name in ["center", "vp_pos", "vp_dir", "width"]:
        param = params[param_name]
        if hasattr(param, "in_units"):
            params[param_name] = param.in_units("code_length")
    bounds = params["bounds"]
    if hasattr(bounds[0], "units"):
        params["bounds"] = tuple(b.in_units("code_length").d for b in bounds)
    return params



import inspect
from textwrap import TextWrapper

import yt

ds = yt.load("RD0005-mine/RedshiftOutput0005")

output = open("source/analyzing/_dq_docstrings.inc", "w")

template = """

.. function:: %(funcname)s%(sig)s:

   (This is a proxy for :func:`~%(funcproxy)s`.)
%(docstring)s

"""

tw = TextWrapper(initial_indent="   ", subsequent_indent="   ", width=60)


def write_docstring(f, name, func):
    docstring = inspect.getdoc(func)
    funcname = name
    sig = inspect.formatargspec(*inspect.getargspec(func))
    sig = sig.replace("data, ", "")
    sig = sig.replace("(data)", "()")
    funcproxy = f"yt.data_objects.derived_quantities.{func.__name__}"
    docstring = "\n".join("   %s" % line for line in docstring.split("\n"))
    f.write(
        template
        % dict(funcname=funcname, sig=sig, funcproxy=funcproxy, docstring=docstring)
    )
    # docstring = "\n".join(tw.wrap(docstring))))


dd = ds.all_data()
for n, func in sorted(dd.quantities.functions.items()):
    print(n, func)
    write_docstring(output, n, func[1])



import numpy as np
from numpy.testing import assert_equal

from yt.testing import assert_rel_equal, fake_random_ds


def setup_module():
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


def test_data_collection():
    # We decompose in different ways
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(16, nprocs=nprocs)
        coll = ds.data_collection(ds.index.grids)
        crho = coll["gas", "density"].sum(dtype="float64").to_ndarray()
        grho = np.sum(
            [g["gas", "density"].sum(dtype="float64") for g in ds.index.grids],
            dtype="float64",
        )
        assert_rel_equal(np.array([crho]), np.array([grho]), 12)
        assert_equal(coll.size, ds.domain_dimensions.prod())
        for gi in range(ds.index.num_grids):
            grids = ds.index.grids[: gi + 1]
            coll = ds.data_collection(grids)
            crho = coll["gas", "density"].sum(dtype="float64")
            grho = np.sum(
                [g["gas", "density"].sum(dtype="float64") for g in grids],
                dtype="float64",
            )
            assert_rel_equal(np.array([crho]), np.array([grho]), 12)
            assert_equal(coll.size, sum(g.ActiveDimensions.prod() for g in grids))



from yt._typing import KnownFieldsT
from yt.fields.field_info_container import FieldInfoContainer

m_units = "code_mass"
p_units = "code_length"
v_units = "code_velocity"

_pnums = 6
_type_fields: KnownFieldsT = tuple(
    ("%s%sType_%d" % (ptype, field, pnum), (units, [], None))
    for pnum in range(_pnums)
    for field, units in (("Mass", m_units), ("Len", p_units))
    for ptype in ("Group", "Subhalo")
)
_sub_type_fields: KnownFieldsT = tuple(
    ("Subhalo%sType_%d" % (field, pnum), (units, [], None))
    for pnum in range(_pnums)
    for field, units in (
        ("HalfmassRad", p_units),
        ("MassInHalfRad", m_units),
        ("MassInMaxRad", m_units),
        ("MassInRad", m_units),
    )
)

_particle_fields: KnownFieldsT = (
    ("GroupPos_0", (p_units, ["Group", "particle_position_x"], None)),
    ("GroupPos_1", (p_units, ["Group", "particle_position_y"], None)),
    ("GroupPos_2", (p_units, ["Group", "particle_position_z"], None)),
    ("GroupVel_0", (v_units, ["Group", "particle_velocity_x"], None)),
    ("GroupVel_1", (v_units, ["Group", "particle_velocity_y"], None)),
    ("GroupVel_2", (v_units, ["Group", "particle_velocity_z"], None)),
    ("GroupMass", (m_units, ["Group", "particle_mass"], None)),
    ("GroupLen", ("", ["Group", "particle_number"], None)),
    ("GroupNsubs", ("", ["Group", "subhalo_number"], None)),
    ("GroupFirstSub", ("", [], None)),
    ("Group_M_Crit200", (m_units, [], None)),
    ("Group_M_Crit500", (m_units, [], None)),
    ("Group_M_Mean200", (m_units, [], None)),
    ("Group_M_TopHat200", (m_units, [], None)),
    ("Group_R_Crit200", (p_units, [], None)),
    ("Group_R_Crit500", (p_units, [], None)),
    ("Group_R_Mean200", (p_units, [], None)),
    ("Group_R_TopHat200", (p_units, [], None)),
    ("SubhaloPos_0", (p_units, ["Subhalo", "particle_position_x"], None)),
    ("SubhaloPos_1", (p_units, ["Subhalo", "particle_position_y"], None)),
    ("SubhaloPos_2", (p_units, ["Subhalo", "particle_position_z"], None)),
    ("SubhaloVel_0", (v_units, ["Subhalo", "particle_velocity_x"], None)),
    ("SubhaloVel_1", (v_units, ["Subhalo", "particle_velocity_y"], None)),
    ("SubhaloVel_2", (v_units, ["Subhalo", "particle_velocity_z"], None)),
    ("SubhaloMass", (m_units, ["Subhalo", "particle_mass"], None)),
    ("SubhaloLen", ("", ["Subhalo", "particle_number"], None)),
    ("SubhaloCM_0", (p_units, ["Subhalo", "center_of_mass_x"], None)),
    ("SubhaloCM_1", (p_units, ["Subhalo", "center_of_mass_y"], None)),
    ("SubhaloCM_2", (p_units, ["Subhalo", "center_of_mass_z"], None)),
    ("SubhaloSpin_0", ("", ["Subhalo", "spin_x"], None)),
    ("SubhaloSpin_1", ("", ["Subhalo", "spin_y"], None)),
    ("SubhaloSpin_2", ("", ["Subhalo", "spin_z"], None)),
    ("SubhaloGrNr", ("", ["Subhalo", "group_identifier"], None)),
    ("SubhaloHalfmassRad", (p_units, [], None)),
    ("SubhaloIDMostbound", ("", [], None)),
    ("SubhaloMassInHalfRad", (m_units, [], None)),
    ("SubhaloMassInMaxRad", (m_units, [], None)),
    ("SubhaloMassInRad", (m_units, [], None)),
    ("SubhaloParent", ("", [], None)),
    ("SubhaloVelDisp", (v_units, ["Subhalo", "velocity_dispersion"], None)),
    ("SubhaloVmax", (v_units, [], None)),
    ("SubhaloVmaxRad", (p_units, [], None)),
    *_type_fields,
    *_sub_type_fields,
)


class GadgetFOFFieldInfo(FieldInfoContainer):
    known_particle_fields = _particle_fields

    # these are extra fields to be created for the "all" particle type
    extra_union_fields = (
        (p_units, "particle_position_x"),
        (p_units, "particle_position_y"),
        (p_units, "particle_position_z"),
        (v_units, "particle_velocity_x"),
        (v_units, "particle_velocity_y"),
        (v_units, "particle_velocity_z"),
        (m_units, "particle_mass"),
        ("", "particle_number"),
        ("", "particle_ones"),
    )


class GadgetFOFHaloFieldInfo(FieldInfoContainer):
    known_particle_fields = _particle_fields + (("ID", ("", ["member_ids"], None)),)



vals = [
    "FluidQuantities",
    "Particles",
    "Parameters",
    "Units",
    "ReadOnDemand",
    "LoadRawData",
    "LevelOfSupport",
    "ContactPerson",
]


class CodeSupport:
    def __init__(self, **kwargs):
        self.support = {}
        for v in vals:
            self.support[v] = "N"
        for k, v in kwargs.items():
            if k in vals:
                self.support[k] = v


Y = "Y"
N = "N"

code_names = ["Enzo", "Orion", "FLASH", "RAMSES", "Chombo", "Gadget", "ART", "ZEUS"]


codes = dict(
    Enzo=CodeSupport(
        FluidQuantities=Y,
        Particles=Y,
        Parameters=Y,
        Units=Y,
        ReadOnDemand=Y,
        LoadRawData=Y,
        ContactPerson="Matt Turk",
        LevelOfSupport="Full",
    ),
    Orion=CodeSupport(
        FluidQuantities=Y,
        Particles=N,
        Parameters=Y,
        Units=Y,
        ReadOnDemand=Y,
        LoadRawData=Y,
        ContactPerson="Jeff Oishi",
        LevelOfSupport="Full",
    ),
    FLASH=CodeSupport(
        FluidQuantities=Y,
        Particles=N,
        Parameters=N,
        Units=Y,
        ReadOnDemand=Y,
        LoadRawData=Y,
        ContactPerson="John !ZuHone",
        LevelOfSupport="Partial",
    ),
    RAMSES=CodeSupport(
        FluidQuantities=Y,
        Particles=N,
        Parameters=N,
        Units=N,
        ReadOnDemand=Y,
        LoadRawData=Y,
        ContactPerson="Matt Turk",
        LevelOfSupport="Partial",
    ),
    Chombo=CodeSupport(
        FluidQuantities=Y,
        Particles=N,
        Parameters=N,
        Units=N,
        ReadOnDemand=Y,
        LoadRawData=Y,
        ContactPerson="Jeff Oishi",
        LevelOfSupport="Partial",
    ),
    Gadget=CodeSupport(
        FluidQuantities=N,
        Particles=Y,
        Parameters=Y,
        Units=Y,
        ReadOnDemand=N,
        LoadRawData=N,
        ContactPerson="Chris Moody",
        LevelOfSupport="Partial",
    ),
    ART=CodeSupport(
        FluidQuantities=N,
        Particles=N,
        Parameters=N,
        Units=N,
        ReadOnDemand=N,
        LoadRawData=N,
        ContactPerson="Matt Turk",
        LevelOfSupport="None",
    ),
    ZEUS=CodeSupport(
        FluidQuantities=N,
        Particles=N,
        Parameters=N,
        Units=N,
        ReadOnDemand=N,
        LoadRawData=N,
        ContactPerson="Matt Turk",
        LevelOfSupport="None",
    ),
)

print("|| . ||", end=" ")
for c in code_names:
    print(f"{c} || ", end=" ")
print()

for vn in vals:
    print(f"|| !{vn} ||", end=" ")
    for c in code_names:
        print(f"{codes[c].support[vn]} || ", end=" ")
    print()



import yt

ds = yt.load("Enzo_64/DD0043/data0043")

sc = yt.create_scene(ds, lens_type="perspective")

source = sc[0]

source.set_field(("gas", "density"))
source.set_log(True)

# Set up the camera parameters: focus, width, resolution, and image orientation
sc.camera.focus = ds.domain_center
sc.camera.resolution = 1024
sc.camera.north_vector = [0, 0, 1]
sc.camera.position = [1.7, 1.7, 1.7]

# You may need to adjust the alpha values to get an image with good contrast.
# For the annotate_domain call, the fourth value in the color tuple is the
# alpha value.
sc.annotate_axes(alpha=0.02)
sc.annotate_domain(ds, color=[1, 1, 1, 0.01])

text_string = f"T = {float(ds.current_time.to('Gyr'))} Gyr"

# save an annotated version of the volume rendering including a representation
# of the transfer function and a nice label showing the simulation time.
sc.save_annotated(
    "vol_annotated.png", sigma_clip=6, text_annotate=[[(0.1, 0.95), text_string]]
)



import yt

# Load the dataset.
ds = yt.load("GasSloshing/sloshing_nomag2_hdf5_plt_cnt_0150")

# Create gas density slices in all three axes.
yt.SlicePlot(ds, "x", ("gas", "density"), width=(800.0, "kpc")).save()
yt.SlicePlot(ds, "y", ("gas", "density"), width=(800.0, "kpc")).save()
yt.SlicePlot(ds, "z", ("gas", "density"), width=(800.0, "kpc")).save()



cdef class OctreeSubsetSelector(SelectorObject):

    def __init__(self, dobj):
        self.base_selector = dobj.base_selector
        self.min_level = self.base_selector.min_level
        self.max_level = self.base_selector.max_level
        self.domain_id = dobj.domain_id
        self.overlap_cells = getattr(dobj.oct_handler, 'overlap_cells', 1)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def select_grids(self,
                     np.ndarray[np.float64_t, ndim=2] left_edges,
                     np.ndarray[np.float64_t, ndim=2] right_edges,
                     np.ndarray[np.int32_t, ndim=2] levels):
        raise RuntimeError

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) noexcept nogil:
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) noexcept nogil:
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_point(self, np.float64_t pos[3]) noexcept nogil:
        return self.base_selector.select_point(pos)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil:
        # return 1
        return self.base_selector.select_bbox(left_edge, right_edge)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_grid(self, np.float64_t left_edge[3],
                         np.float64_t right_edge[3], np.int32_t level,
                         Oct *o = NULL) noexcept nogil:
        # Because visitors now use select_grid, we should be explicitly
        # checking this.
        cdef int res
        res = self.base_selector.select_grid(left_edge, right_edge, level, o)
        if self.domain_id == -1:
            return res
        elif res == 1 and o != NULL and o.domain != self.domain_id:
            return -1
        return res

    def _hash_vals(self):
        return (hash(self.base_selector), self.domain_id)

octree_subset_selector = OctreeSubsetSelector



cdef class IndexedOctreeSubsetSelector(SelectorObject):
    # This is a numpy array, which will be a bool of ndim 1
    cdef np.uint64_t min_ind
    cdef np.uint64_t max_ind
    cdef public SelectorObject base_selector
    cdef int filter_bbox
    cdef np.float64_t DLE[3]
    cdef np.float64_t DRE[3]

    def __init__(self, dobj):
        self.min_ind = dobj.min_ind
        self.max_ind = dobj.max_ind
        self.base_selector = dobj.base_selector
        self.min_level = self.base_selector.min_level
        self.max_level = self.base_selector.max_level
        self.filter_bbox = 0
        if getattr(dobj.ds, "filter_bbox", False):
            self.filter_bbox = 1
        for i in range(3):
            self.DLE[i] = dobj.ds.domain_left_edge[i]
            self.DRE[i] = dobj.ds.domain_right_edge[i]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def select_grids(self,
                     np.ndarray[np.float64_t, ndim=2] left_edges,
                     np.ndarray[np.float64_t, ndim=2] right_edges,
                     np.ndarray[np.int32_t, ndim=2] levels):
        raise RuntimeError

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) noexcept nogil:
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) noexcept nogil:
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_point(self, np.float64_t pos[3]) noexcept nogil:
        cdef int i
        if self.filter_bbox == 0:
            return 1
        for i in range(3):
            if pos[i] < self.DLE[i] or pos[i] > self.DRE[i]:
                return 0
        return 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil:
        return self.base_selector.select_bbox(left_edge, right_edge)

    cdef int select_grid(self, np.float64_t left_edge[3],
                         np.float64_t right_edge[3], np.int32_t level,
                         Oct *o = NULL) noexcept nogil:
        # Because visitors now use select_grid, we should be explicitly
        # checking this.
        return self.base_selector.select_grid(left_edge, right_edge, level, o)

    def _hash_vals(self):
        return (hash(self.base_selector), self.min_ind, self.max_ind)

indexed_octree_subset_selector = IndexedOctreeSubsetSelector



from collections import defaultdict
from functools import lru_cache
from typing import TYPE_CHECKING, Union

import numpy as np
from unyt import unyt_array

from yt._maintenance.deprecation import issue_deprecation_warning
from yt.frontends.ramses.definitions import VAR_DESC_RE, VERSION_RE
from yt.utilities.cython_fortran_utils import FortranFile
from yt.utilities.exceptions import (
    YTFieldTypeNotFound,
    YTFileNotParseable,
    YTParticleOutputFormatNotImplemented,
)
from yt.utilities.io_handler import BaseIOHandler
from yt.utilities.logger import ytLogger as mylog

if TYPE_CHECKING:
    import os


def convert_ramses_ages(ds, conformal_ages):
    issue_deprecation_warning(
        msg=(
            "The `convert_ramses_ages' function is deprecated. It should be replaced "
            "by the `convert_ramses_conformal_time_to_physical_age' function."
        ),
        stacklevel=3,
        since="4.0.3",
    )
    return convert_ramses_conformal_time_to_physical_time(ds, conformal_ages)


def convert_ramses_conformal_time_to_physical_time(
    ds, conformal_time: np.ndarray
) -> unyt_array:
    """
    Convert conformal times (as defined in RAMSES) to physical times.

    Arguments
    ---------
    ds : RAMSESDataset
        The RAMSES dataset to use for the conversion
    conformal_time : np.ndarray
        The conformal time as read from disk

    Returns
    -------
    physical_age : np.ndarray
        The physical age in code units
    """
    h0 = ds.hubble_constant
    tau_bins = ds.tau_frw * h0
    t_bins = ds.t_frw

    min_time = 0
    max_time = ds.current_time.to(t_bins.units)

    return ds.arr(
        np.clip(
            np.interp(
                conformal_time,
                tau_bins,
                t_bins.value,
                right=max_time,
                left=min_time,
            ),
            min_time,
            max_time.value,
        ),
        t_bins.units,
    )


def _ramses_particle_binary_file_handler(particle_handler, subset, fields, count):
    """General file handler for binary file, called by _read_particle_subset

    Parameters
    ----------
    particle : ``ParticleFileHandler``
        the particle class we want to read
    subset: ``RAMSESDomainSubset``
        A RAMSES domain subset object
    fields: list of tuple
        The fields to read
    count: integer
        The number of elements to count
    """
    tr = {}
    ds = subset.domain.ds
    foffsets = particle_handler.field_offsets
    fname = particle_handler.fname
    data_types = particle_handler.field_types
    with FortranFile(fname) as fd:
        # We do *all* conversion into boxlen here.
        # This means that no other conversions need to be applied to convert
        # positions into the same domain as the octs themselves.
        for field in sorted(fields, key=lambda a: foffsets[a]):
            if count == 0:
                tr[field] = np.empty(0, dtype=data_types[field])
                continue
            # Sentinel value: -1 means we don't have this field
            if foffsets[field] == -1:
                tr[field] = np.empty(count, dtype=data_types[field])
            else:
                fd.seek(foffsets[field])
                dt = data_types[field]
                tr[field] = fd.read_vector(dt)
            if field[1].startswith("particle_position"):
                np.divide(tr[field], ds["boxlen"], tr[field])

            # Hand over to field handler for special cases, like particle_birth_times
            particle_handler.handle_field(field, tr)

    return tr


def _ramses_particle_csv_file_handler(particle_handler, subset, fields, count):
    """General file handler for csv file, called by _read_particle_subset

    Parameters
    ----------
    particle: ``ParticleFileHandler``
        the particle class we want to read
    subset: ``RAMSESDomainSubset``
        A RAMSES domain subset object
    fields: list of tuple
        The fields to read
    count: integer
        The number of elements to count
    """
    from yt.utilities.on_demand_imports import _pandas as pd

    tr = {}
    ds = subset.domain.ds
    foffsets = particle_handler.field_offsets
    fname = particle_handler.fname

    list_field_ind = [
        (field, foffsets[field]) for field in sorted(fields, key=lambda a: foffsets[a])
    ]

    # read only selected fields
    dat = pd.read_csv(
        fname,
        delimiter=",",
        usecols=[ind for _field, ind in list_field_ind],
        skiprows=2,
        header=None,
    )

    for field, ind in list_field_ind:
        tr[field] = dat[ind].to_numpy()
        if field[1].startswith("particle_position"):
            np.divide(tr[field], ds["boxlen"], tr[field])

        particle_handler.handle_field(field, tr)

    return tr


class IOHandlerRAMSES(BaseIOHandler):
    _dataset_type = "ramses"

    def _read_fluid_selection(self, chunks, selector, fields, size):
        tr = defaultdict(list)

        # Set of field types
        ftypes = {f[0] for f in fields}
        for chunk in chunks:
            # Gather fields by type to minimize i/o operations
            for ft in ftypes:
                # Get all the fields of the same type
                field_subs = list(filter(lambda f, ft=ft: f[0] == ft, fields))

                # Loop over subsets
                for subset in chunk.objs:
                    fname = None
                    for fh in subset.domain.field_handlers:
                        if fh.ftype == ft:
                            file_handler = fh
                            fname = fh.fname
                            break

                    if fname is None:
                        raise YTFieldTypeNotFound(ft)

                    # Now we read the entire thing
                    with FortranFile(fname) as fd:
                        # This contains the boundary information, so we skim through
                        # and pick off the right vectors
                        rv = subset.fill(fd, field_subs, selector, file_handler)
                    for ft, f in field_subs:
                        d = rv.pop(f)
                        if d.size == 0:
                            continue
                        mylog.debug(
                            "Filling %s with %s (%0.3e %0.3e) (%s zones)",
                            f,
                            d.size,
                            d.min(),
                            d.max(),
                            d.size,
                        )
                        tr[ft, f].append(d)
        d = {}
        for field in fields:
            tmp = tr.pop(field, None)
            d[field] = np.concatenate(tmp) if tmp else np.empty(0, dtype="d")

        return d

    def _read_particle_coords(self, chunks, ptf):
        pn = "particle_position_%s"
        fields = [
            (ptype, f"particle_position_{ax}")
            for ptype, field_list in ptf.items()
            for ax in "xyz"
        ]
        for chunk in chunks:
            for subset in chunk.objs:
                rv = self._read_particle_subset(subset, fields)
                for ptype in sorted(ptf):
                    yield (
                        ptype,
                        (
                            rv[ptype, pn % "x"],
                            rv[ptype, pn % "y"],
                            rv[ptype, pn % "z"],
                        ),
                        0.0,
                    )

    def _read_particle_fields(self, chunks, ptf, selector):
        pn = "particle_position_%s"
        chunks = list(chunks)
        fields = [
            (ptype, fname) for ptype, field_list in ptf.items() for fname in field_list
        ]

        for ptype, field_list in sorted(ptf.items()):
            for ax in "xyz":
                if pn % ax not in field_list:
                    fields.append((ptype, pn % ax))

            if ptype == "sink_csv":
                subset = chunks[0].objs[0]
                rv = self._read_particle_subset(subset, fields)
                for ptype, field_list in sorted(ptf.items()):
                    x, y, z = (np.asarray(rv[ptype, pn % ax], "=f8") for ax in "xyz")
                    mask = selector.select_points(x, y, z, 0.0)
                    if mask is None:
                        mask = []
                    for field in field_list:
                        data = np.asarray(rv.pop((ptype, field))[mask], "=f8")
                        yield (ptype, field), data

            else:
                for chunk in chunks:
                    for subset in chunk.objs:
                        rv = self._read_particle_subset(subset, fields)
                        for ptype, field_list in sorted(ptf.items()):
                            x, y, z = (
                                np.asarray(rv[ptype, pn % ax], "=f8") for ax in "xyz"
                            )
                            mask = selector.select_points(x, y, z, 0.0)
                            if mask is None:
                                mask = []
                            for field in field_list:
                                data = np.asarray(rv.pop((ptype, field))[mask], "=f8")
                                yield (ptype, field), data

    def _read_particle_subset(self, subset, fields):
        """Read the particle files."""
        tr = {}

        # Sequential read depending on particle type
        for ptype in {f[0] for f in fields}:
            # Select relevant files
            subs_fields = filter(lambda f, ptype=ptype: f[0] == ptype, fields)

            ok = False
            for ph in subset.domain.particle_handlers:
                if ph.ptype == ptype:
                    ok = True
                    count = ph.local_particle_count
                    break
            if not ok:
                raise YTFieldTypeNotFound(ptype)

            tr.update(ph.reader(subset, subs_fields, count))

        return tr


@lru_cache
def _read_part_binary_file_descriptor(fname: Union[str, "os.PathLike[str]"]):
    """
    Read a file descriptor and returns the array of the fields found.
    """

    # Mapping
    mapping_list = [
        ("position_x", "particle_position_x"),
        ("position_y", "particle_position_y"),
        ("position_z", "particle_position_z"),
        ("velocity_x", "particle_velocity_x"),
        ("velocity_y", "particle_velocity_y"),
        ("velocity_z", "particle_velocity_z"),
        ("mass", "particle_mass"),
        ("identity", "particle_identity"),
        ("levelp", "particle_level"),
        ("family", "particle_family"),
        ("tag", "particle_tag"),
    ]
    # Convert to dictionary
    mapping = dict(mapping_list)

    with open(fname) as f:
        line = f.readline()
        tmp = VERSION_RE.match(line)
        mylog.debug("Reading part file descriptor %s.", fname)
        if not tmp:
            raise YTParticleOutputFormatNotImplemented()

        version = int(tmp.group(1))

        if version == 1:
            # Skip one line (containing the headers)
            line = f.readline()
            fields = []
            for i, line in enumerate(f.readlines()):
                tmp = VAR_DESC_RE.match(line)
                if not tmp:
                    raise YTFileNotParseable(fname, i + 1)

                # ivar = tmp.group(1)
                varname = tmp.group(2)
                dtype = tmp.group(3)

                if varname in mapping:
                    varname = mapping[varname]
                else:
                    varname = f"particle_{varname}"

                fields.append((varname, dtype))
        else:
            raise YTParticleOutputFormatNotImplemented()

    return fields


@lru_cache
def _read_part_csv_file_descriptor(fname: Union[str, "os.PathLike[str]"]):
    """
    Read the file from the csv sink particles output.
    """
    from yt.utilities.on_demand_imports import _pandas as pd

    # Fields name from the default csv RAMSES sink algorithm in the yt default convention
    mapping = {
        " # id": "particle_identifier",
        "msink": "particle_mass",
        "x": "particle_position_x",
        "y": "particle_position_y",
        "z": "particle_position_z",
        "vx": "particle_velocity_x",
        "vy": "particle_velocity_y",
        "vz": "particle_velocity_z",
        "lx": "particle_angular_momentum_x",
        "ly": "particle_angular_momentum_y",
        "lz": "particle_angular_momentum_z",
        "tform": "particle_formation_time",
        "acc_rate": "particle_accretion_rate",
        "del_mass": "particle_delta_mass",
        "rho_gas": "particle_rho_gas",
        "cs**2": "particle_sound_speed",
        "etherm": "particle_etherm",
        "vx_gas": "particle_velocity_x_gas",
        "vy_gas": "particle_velocity_y_gas",
        "vz_gas": "particle_velocity_z_gas",
        "mbh": "particle_mass_bh",
        "level": "particle_level",
        "rsink_star": "particle_radius_star",
    }

    # read the all file to get the number of particle
    dat = pd.read_csv(fname, delimiter=",")
    fields = []
    local_particle_count = len(dat)

    for varname in dat.columns:
        if varname in mapping:
            varname = mapping[varname]
        else:
            varname = f"particle_{varname}"

        fields.append(varname)

    return fields, local_particle_count


@lru_cache
def _read_fluid_file_descriptor(fname: Union[str, "os.PathLike[str]"], *, prefix: str):
    """
    Read a file descriptor and returns the array of the fields found.
    """

    # Mapping
    mapping_list = [
        ("density", "Density"),
        ("velocity_x", "x-velocity"),
        ("velocity_y", "y-velocity"),
        ("velocity_z", "z-velocity"),
        ("pressure", "Pressure"),
        ("metallicity", "Metallicity"),
        # Add mapping for ionized species
        # Note: we expect internally that these names use the HII, HeII,
        #       HeIII, ... convention for historical reasons. So we need to map
        #       the names read from `hydro_file_descriptor.txt` to this
        #       convention.
        # This will create fields like ("ramses", "HII") which are mapped
        # to ("gas", "H_p1_fraction") in fields.py
        ("H_p1_fraction", "HII"),
        ("He_p1_fraction", "HeII"),
        ("He_p2_fraction", "HeIII"),
        # Photon fluxes / densities are stored as `photon_density_XX`, so
        # only 100 photon bands can be stored with this format. Let's be
        # conservative and support up to 100 bands.
        *[(f"photon_density_{i:02d}", f"Photon_density_{i:d}") for i in range(100)],
        *[
            (f"photon_flux_{i:02d}_{dim}", f"Photon_flux_{dim}_{i:d}")
            for i in range(100)
            for dim in "xyz"
        ],
    ]

    # Add mapping for magnetic fields
    mapping_list += [
        (key, key)
        for key in (
            f"B_{dim}_{side}" for side in ["left", "right"] for dim in ["x", "y", "z"]
        )
    ]

    # Convert to dictionary
    mapping = dict(mapping_list)

    with open(fname) as f:
        line = f.readline()
        tmp = VERSION_RE.match(line)
        mylog.debug("Reading fluid file descriptor %s.", fname)
        if not tmp:
            return []

        version = int(tmp.group(1))

        if version == 1:
            # Skip one line (containing the headers)
            line = f.readline()
            fields = []
            for i, line in enumerate(f.readlines()):
                tmp = VAR_DESC_RE.match(line)
                if not tmp:
                    raise YTFileNotParseable(fname, i + 1)

                # ivar = tmp.group(1)
                varname = tmp.group(2)
                dtype = tmp.group(3)

                if varname in mapping:
                    varname = mapping[varname]
                else:
                    varname = f"{prefix}_{varname}"

                fields.append((varname, dtype))
        else:
            mylog.error("Version %s", version)
            raise YTParticleOutputFormatNotImplemented()

    return fields



"""
Definitions for image samplers




"""


import numpy as np

cimport cython
cimport numpy as np

from .partitioned_grid cimport PartitionedGrid
from .volume_container cimport VolumeContainer

cdef enum:
    Nch = 4

# NOTE: We don't want to import the field_interpolator_tables here, as it
# breaks a bunch of C++ interop.  Maybe some day it won't.  So, we just forward
# declare.
cdef struct VolumeRenderAccumulator

ctypedef int calculate_extent_function(ImageSampler image,
            VolumeContainer *vc, np.int64_t rv[4]) except -1 nogil

ctypedef void generate_vector_info_function(ImageSampler im,
            np.int64_t vi, np.int64_t vj,
            np.float64_t width[2],
            np.float64_t v_dir[3], np.float64_t v_pos[3]) noexcept nogil

cdef struct ImageAccumulator:
    np.float64_t rgba[Nch]
    void *supp_data

cdef class ImageSampler:
    cdef np.float64_t[:,:,:] vp_pos
    cdef np.float64_t[:,:,:] vp_dir
    cdef np.float64_t *center
    cdef np.float64_t[:,:,:] image
    cdef np.float64_t[:,:] zbuffer
    cdef np.int64_t[:,:] image_used
    cdef np.int64_t[:,:] mesh_lines
    cdef np.float64_t pdx, pdy
    cdef np.float64_t bounds[4]
    cdef np.float64_t[:,:] camera_data   # position, width, unit_vec[0,2]
    cdef int nv[2]
    cdef np.float64_t *x_vec
    cdef np.float64_t *y_vec
    cdef public object acenter, aimage, ax_vec, ay_vec
    cdef public object azbuffer
    cdef public object aimage_used
    cdef public object amesh_lines
    cdef void *supp_data
    cdef np.float64_t width[3]
    cdef public object lens_type
    cdef public str volume_method
    cdef calculate_extent_function *extent_function
    cdef generate_vector_info_function *vector_function
    cdef void setup(self, PartitionedGrid pg)
    @staticmethod
    cdef void sample(VolumeContainer *vc,
                np.float64_t v_pos[3],
                np.float64_t v_dir[3],
                np.float64_t enter_t,
                np.float64_t exit_t,
                int index[3],
                void *data) noexcept nogil

cdef class ProjectionSampler(ImageSampler):
    pass

cdef class InterpolatedProjectionSampler(ImageSampler):
    cdef VolumeRenderAccumulator *vra
    cdef public object tf_obj
    cdef public object my_field_tables

cdef class VolumeRenderSampler(ImageSampler):
    cdef VolumeRenderAccumulator *vra
    cdef public object tf_obj
    cdef public object my_field_tables
    cdef object tree_containers

cdef class LightSourceRenderSampler(ImageSampler):
    cdef VolumeRenderAccumulator *vra
    cdef public object tf_obj
    cdef public object my_field_tables
